#!/usr/bin/env python
# crate_anon/crateweb/research/research_db_info.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

from collections import OrderedDict
# from functools import lru_cache
import logging
from typing import Any, Dict, List, Optional

from django.db import connections
from django.db.backends.base.base import BaseDatabaseWrapper
from django.conf import settings
from openpyxl import Workbook

from crate_anon.common.jsonfunc import METHOD_NO_ARGS, register_for_json
from crate_anon.common.sql import (
    ColumnId,
    is_sql_column_type_textual,
    make_grammar,
    QB_DATATYPE_DATE,
    QB_DATATYPE_FLOAT,
    QB_DATATYPE_INTEGER,
    QB_DATATYPE_STRING,
    QB_DATATYPE_STRING_FULLTEXT,
    QB_DATATYPE_UNKNOWN,
    SchemaId,
    SqlArgsTupleType,
    SQLTYPES_FLOAT,
    SQLTYPES_WITH_DATE,
    SQLTYPES_TEXT,
    SQLTYPES_INTEGER,
    TableId,
    translate_sql_qmark_to_percent,
)
from crate_anon.common.sqla import (
    MSSQL_DEFAULT_SCHEMA,
    POSTGRES_DEFAULT_SCHEMA,
)
from crate_anon.common.sql_grammar import SqlGrammar
from crate_anon.common.sql_grammar_factory import (
    DIALECT_MSSQL,
    DIALECT_POSTGRES,
    DIALECT_MYSQL,
)
from crate_anon.crateweb.core.dbfunc import (
    dictfetchall,
    dictlist_to_tsv,
)
from crate_anon.crateweb.extra.django_cache_fn import django_cache_function
from crate_anon.crateweb.extra.excel import excel_to_bytes

log = logging.getLogger(__name__)


# =============================================================================
# Constants
# =============================================================================

# https://docs.djangoproject.com/en/1.10/howto/custom-lookups/#writing-alternative-implementations-for-existing-lookups  # noqa
VENDOR_MICROSOFT = 'microsoft'
VENDOR_POSTGRESQL = 'postgresql'
VENDOR_MYSQL = 'mysql'


# =============================================================================
# Information about the research database
# =============================================================================

class ColumnInfo(object):
    # See also querybuilder.js

    def __init__(self, **kwargs) -> None:
        self.table_catalog = kwargs.pop('table_catalog')
        self.table_schema = kwargs.pop('table_schema')
        self.table_name = kwargs.pop('table_name')
        self.column_name = kwargs.pop('column_name')
        self.is_nullable = bool(kwargs.pop('is_nullable'))
        self.column_type = kwargs.pop('column_type')
        self.column_comment = kwargs.pop('column_comment')
        self.indexed = bool(kwargs.pop('indexed'))
        self.indexed_fulltext = bool(kwargs.pop('indexed_fulltext'))

    def basetype(self) -> str:
        return self.column_type.split("(")[0].upper()

    def querybuilder_type(self) -> str:
        """
        Returns a string that is defined in querybuilder.js
        """
        basetype = self.basetype()
        if basetype in SQLTYPES_FLOAT:
            return QB_DATATYPE_FLOAT
        if basetype in SQLTYPES_WITH_DATE:
            return QB_DATATYPE_DATE
        if basetype in SQLTYPES_TEXT:
            if self.indexed_fulltext:
                return QB_DATATYPE_STRING_FULLTEXT
            else:
                return QB_DATATYPE_STRING
        if basetype in SQLTYPES_INTEGER:
            return QB_DATATYPE_INTEGER
        return QB_DATATYPE_UNKNOWN

    def column_id(self) -> ColumnId:
        return ColumnId(db=self.table_catalog,
                        schema=self.table_schema,
                        table=self.table_name,
                        column=self.column_name)

    def table_id(self) -> TableId:
        return TableId(db=self.table_catalog,
                       schema=self.table_schema,
                       table=self.table_name)


@register_for_json(method=METHOD_NO_ARGS)
class ResearchDatabaseInfo(object):
    """
    Fetches schema information from the research database.
    Class primarily exists to be able to use @cached_property.
    ... replaced by @lru_cache
    ... replaced by @django_cache_function
    """
    # We fetch the dialect at first request; this enables us to import the
    # class without Django configured.

    def __init__(self) -> None:
        self._dialect = None
        self._grammar = None

    @property
    def dialect(self) -> str:
        if not self._dialect:
            self._dialect = settings.RESEARCH_DB_DIALECT
        return self._dialect

    @property
    def grammar(self) -> SqlGrammar:
        if not self._grammar:
            self._grammar = make_grammar(self.dialect)
        return self._grammar

    @classmethod
    def _connection(cls) -> BaseDatabaseWrapper:
        return connections['research']

    @classmethod
    def uses_database_level(cls) -> bool:
        return cls._offers_db_above_schema(cls._connection())

    @classmethod
    def format_db_schema(cls, db: str, schema: str) -> str:
        if cls.uses_database_level():
            return "{}.{}".format(db, schema)
        else:
            return schema

    @staticmethod
    def _offers_db_above_schema(connection: BaseDatabaseWrapper) -> bool:
        return connection.vendor in [VENDOR_MICROSOFT]
        # not MySQL ("database" concept = "schema" concept)
        # not PostgreSQL (only one database per connection)

    @staticmethod
    def get_researchdb_schemas() -> List[SchemaId]:
        return [SchemaId(db=x['database'], schema=x['schema'])
                for x in settings.RESEARCH_DB_INFO]

    def get_default_database_name(self) -> str:
        dialect = self.dialect
        if dialect == DIALECT_MSSQL:
            return settings.DATABASES['research']['NAME']
        elif dialect == DIALECT_POSTGRES:
            return ''
        elif dialect == DIALECT_MYSQL:
            return ''
        else:
            raise ValueError("Bad settings.RESEARCH_DB_DIALECT")

    def get_default_schema_name(self) -> str:
        dialect = self.dialect
        if dialect == DIALECT_MSSQL:
            return MSSQL_DEFAULT_SCHEMA
        elif dialect == DIALECT_POSTGRES:
            return POSTGRES_DEFAULT_SCHEMA
        elif dialect == DIALECT_MYSQL:
            return settings.DATABASES['research']['NAME']
        else:
            raise ValueError("Bad settings.RESEARCH_DB_DIALECT")

    def get_db_info(self, schema: SchemaId) -> Optional[Dict[str, Any]]:
        db_name = schema.db() or self.get_default_database_name()
        schema_name = schema.schema() or self.get_default_schema_name()
        for x in settings.RESEARCH_DB_INFO:
            if x['database'] == db_name and x['schema'] == schema_name:
                return x
        log.warning("No such database/schema: {}".format(
            schema.identifier(self.grammar)))
        return None

    def get_schema_trid_field(self, schema: SchemaId) -> str:
        db_info = self.get_db_info(schema)
        if not db_info:
            return ''
        return db_info.get('trid_field', '')

    def get_schema_rid_field(self, schema: SchemaId) -> str:
        schema_info = self.get_db_info(schema)
        if not schema_info:
            return ''
        return schema_info.get('rid_field', '')

    def get_db_rid_family(self, schema: SchemaId) -> str:
        db_info = self.get_db_info(schema)
        if not db_info:
            return ''
        return db_info.get('rid_family', '')

    def get_db_mrid_table(self, schema: SchemaId) -> str:
        db_info = self.get_db_info(schema)
        if not db_info:
            return ''
        return db_info.get('mrid_table', '')

    def get_db_mrid_field(self, schema: SchemaId) -> str:
        db_info = self.get_db_info(schema)
        if not db_info:
            return ''
        return db_info.get('mrid_field', '')

    def get_schema_date_field(self, schema: SchemaId) -> str:
        db_info = self.get_db_info(schema)
        if not db_info:
            return ''
        return db_info.get('default_date_field', '')

    def get_rid_column(self, table: TableId) -> ColumnId:
        # RID column in the specified table (which may or may not exist)
        return table.column_id(self.get_schema_rid_field(table.schema_id()))

    def get_trid_column(self, table: TableId) -> ColumnId:
        # TRID column in the specified table (which may or may not exist)
        return table.column_id(
            self.get_schema_trid_field(table.schema_id()))

    def get_mrid_column_from_schema(self, schema: SchemaId) -> ColumnId:
        # MRID column in the MRID master table
        return schema.column_id(
            table=self.get_db_mrid_table(schema),
            column=self.get_db_mrid_field(schema))

    def get_mrid_column_from_table(self, table: TableId) -> ColumnId:
        # MRID column in the MRID master table
        return self.get_mrid_column_from_schema(table.schema_id())

    def get_linked_mrid_column(self, table: TableId) -> Optional[ColumnId]:
        """
        Returns either the MRID column in the schema containing the table
        specified, or one that can be linked to it automatically.
        """
        own_mrid = self.get_mrid_column_from_table(table)
        if own_mrid:
            return own_mrid
        # OK. So our table isn't from a database with an MRID table, but it
        # might be linked to one.
        this_schema = table.schema_id()
        first_schema = self.get_first_schema()
        if not first_schema or not self.talks_to_world(first_schema):
            return None
        if self.talks_to_world(this_schema) or self.can_communicate_directly(
                this_schema, first_schema):
            return self.get_mrid_column_from_schema(first_schema)

    def get_default_date_column(self, table: TableId) -> ColumnId:
        # Default date column (which may or may not exist)
        return table.column_id(
            self.get_schema_date_field(table.schema_id()))

    def does_db_schema_have_mrid(self, schema: SchemaId) -> bool:
        this_dbs_info = self.get_db_info(schema)
        if not this_dbs_info:
            return False
        return bool(
            this_dbs_info.get('mrid_table', None) and
            this_dbs_info.get('mrid_field', None)
        )

    @staticmethod
    def get_first_schema() -> SchemaId:
        if not settings.RESEARCH_DB_INFO:
            return SchemaId()
        first_dbs_info = settings.RESEARCH_DB_INFO[0]
        return SchemaId(db=first_dbs_info['database'],
                        schema=first_dbs_info['schema'])

    def talks_to_world(self, schema: SchemaId) -> bool:
        return self.does_db_schema_have_mrid(schema)

    def can_communicate_directly(self,
                                 schema1: SchemaId,
                                 schema2: SchemaId) -> bool:
        if schema1 == schema2:
            return True
        info1 = self.get_db_info(schema1)
        info2 = self.get_db_info(schema2)
        if not info1 or not info2:
            return False
        return bool(
            info1.get('rid_field', None) and info1.get('rid_family', None) and
            info2.get('rid_field', None) and info2.get('rid_family', None) and
            info2.get('rid_family', None) == info1.get('rid_family', None)
        )

    def is_db_schema_eligible_for_query_builder(self,
                                                schema: SchemaId) -> bool:
        first_schema = self.get_first_schema()
        if not schema or not first_schema:
            return False
        if schema == first_schema:
            # First one: always eligible
            return True
        return (
            (self.talks_to_world(first_schema) and
             self.talks_to_world(schema)) or
            self.can_communicate_directly(schema, first_schema)
        )

    @classmethod
    def _get_info_microsoft(cls, db_name: str,
                            schema_names: List[str]) -> SqlArgsTupleType:
        if not schema_names:
            raise ValueError("No schema_names specified (for SQL Server "
                             "database)")
        # SQL Server INFORMATION_SCHEMA.COLUMNS:
        # - https://msdn.microsoft.com/en-us/library/ms188348.aspx
        # Re fulltext indexes:
        # - http://stackoverflow.com/questions/16280918/how-to-find-full-text-indexing-on-database-in-sql-server-2008  # noqa
        # - sys.fulltext_indexes: https://msdn.microsoft.com/en-us/library/ms186903.aspx  # noqa
        # - sys.fulltext_catalogs: https://msdn.microsoft.com/en-us/library/ms188779.aspx  # noqa
        # - sys.fulltext_index_columns: https://msdn.microsoft.com/en-us/library/ms188335.aspx  # noqa
        schema_placeholder = ",".join(["?"] * len(schema_names))
        sql = translate_sql_qmark_to_percent("""
SELECT
    ? AS table_catalog,
    d.table_schema,
    d.table_name,
    d.column_name,
    d.is_nullable,
    d.column_type,
    d.column_comment,
    CASE WHEN COUNT(d.index_id) > 0 THEN 1 ELSE 0 END AS indexed,
    CASE WHEN COUNT(d.fulltext_index_object_id) > 0 THEN 1 ELSE 0 END AS indexed_fulltext
FROM (
    SELECT
        s.name AS table_schema,
        ta.name AS table_name,
        c.name AS column_name,
        c.is_nullable,
        UPPER(ty.name) + '(' + CONVERT(VARCHAR(100), c.max_length) + ')' AS column_type,
        CONVERT(VARCHAR(1000), x.value) AS column_comment, -- x.value is of type SQL_VARIANT
        i.index_id,
        fi.object_id AS fulltext_index_object_id
    FROM [{db_name}].sys.tables ta
    INNER JOIN [{db_name}].sys.schemas s ON ta.schema_id = s.schema_id
    INNER JOIN [{db_name}].sys.columns c ON c.object_id = ta.object_id
    INNER JOIN [{db_name}].sys.types ty ON ty.system_type_id = c.system_type_id
    LEFT JOIN [{db_name}].sys.extended_properties x ON (
        x.major_id = c.object_id
        AND x.minor_id = c.column_id
    )
    LEFT JOIN [{db_name}].sys.index_columns i ON (
        i.object_id = c.object_id
        AND i.column_id = c.column_id
    )
    LEFT JOIN [{db_name}].sys.fulltext_index_columns fi ON (
        fi.object_id = c.object_id
        AND fi.column_id = c.column_id
    )
    WHERE s.name IN ({schema_placeholder})
    AND ty.user_type_id = ty.system_type_id  -- restricts to system data types; eliminates 'sysname' type
) AS d
GROUP BY
    table_schema,
    table_name,
    column_name,
    is_nullable,
    column_type,
    column_comment
ORDER BY
    table_schema,
    table_name,
    column_name
        """.format(db_name=db_name,  # noqa
                   schema_placeholder=schema_placeholder))
        args = [db_name] + schema_names
        return sql, args

    @classmethod
    def _get_info_mysql(cls, db_and_schema_name: str) -> SqlArgsTupleType:
        # ---------------------------------------------------------------------
        # Method A. Stupidly slow, e.g. 47s for the query.
        # ---------------------------------------------------------------------
        # It's the EXISTS stuff that's slow.
        #
        # sql = translate_sql_qmark_to_percent("""
        #     SELECT
        #         c.table_schema,
        #         c.table_name,
        #         c.column_name,
        #         c.is_nullable,
        #         c.column_type,  /* MySQL: e.g. varchar(32) */
        #         c.column_comment,  /* MySQL */
        #         EXISTS (
        #             SELECT *
        #             FROM information_schema.statistics s
        #             WHERE s.table_schema = c.table_schema
        #             AND s.table_name = c.table_name
        #             AND s.column_name = c.column_name
        #         ) AS indexed,
        #         EXISTS (
        #             SELECT *
        #             FROM information_schema.statistics s
        #             WHERE s.table_schema = c.table_schema
        #             AND s.table_name = c.table_name
        #             AND s.column_name = c.column_name
        #             AND s.index_type LIKE 'FULLTEXT%'
        #         ) AS indexed_fulltext
        #     FROM
        #         information_schema.columns c
        #     WHERE
        #         c.table_schema IN ({schema_placeholder})
        #     ORDER BY
        #         c.table_schema,
        #         c.table_name,
        #         c.column_name
        # """.format(
        #     schema_placeholder=",".join(["?"] * len(schemas)),
        # ))
        # args = schemas
        #
        # ---------------------------------------------------------------------
        # Method B. Much faster, e.g. 0.35s for the same thing.
        # ---------------------------------------------------------------------
        # http://www.codeproject.com/Articles/33052/Visual-Representation-of-SQL-Joins  # noqa
        # (Note that EXISTS() above returns 0 or 1.)
        # The LEFT JOIN below will produce NULL values for the index
        # columns for non-indexed fields.
        # However, you can have more than one index on a column, in which
        # case the column appears in two rows.
        #
        # MySQL's INFORMATION_SCHEMA.COLUMNS:
        # - https://dev.mysql.com/doc/refman/5.7/en/tables-table.html
        sql = translate_sql_qmark_to_percent("""
SELECT
    '' AS table_catalog,
    d.table_schema,
    d.table_name,
    d.column_name,
    d.is_nullable,
    d.column_type,
    d.column_comment,
    d.indexed,
    MAX(d.indexed_fulltext) AS indexed_fulltext
FROM (
    SELECT
        -- c.table_catalog,  -- will always be 'def'
        c.table_schema,
        c.table_name,
        c.column_name,
        c.is_nullable,
        c.column_type,  /* MySQL: e.g. varchar(32) */
        c.column_comment,  /* MySQL */
        /* s.index_name, */
        /* s.index_type, */
        IF(s.index_type IS NOT NULL, 1, 0) AS indexed,
        IF(s.index_type LIKE 'FULLTEXT%', 1, 0) AS indexed_fulltext
    FROM
        information_schema.columns c
        LEFT JOIN information_schema.statistics s
        ON (
            c.table_schema = s.table_schema
            AND c.table_name = s.table_name
            AND c.column_name = s.column_name
        )
    WHERE
        c.table_schema = ?
) AS d  /* "Every derived table must have its own alias" */
GROUP BY
    table_catalog,
    table_schema,
    table_name,
    column_name,
    is_nullable,
    column_type,
    column_comment,
    indexed
ORDER BY
    table_catalog,
    table_schema,
    table_name,
    column_name
        """)
        args = [db_and_schema_name]
        return sql, args

    @classmethod
    def _get_info_postgres(cls, schema_names: List[str]) -> SqlArgsTupleType:
        # A PostgreSQL connection is always to a single database.
        # http://stackoverflow.com/questions/10335561/use-database-name-command-in-postgresql  # noqa
        if not schema_names:
            raise ValueError("No schema_names specified (for PostgreSQL "
                             "database)")
        # http://dba.stackexchange.com/questions/75015
        # http://stackoverflow.com/questions/14713774
        # Note that creating a GIN index looks like:
        #       ALTER TABLE t ADD COLUMN tsv_mytext TSVECTOR;
        #       UPDATE t SET tsv_mytext = to_tsvector(mytext);
        #       CREATE INDEX idx_t_mytext_gin ON t USING GIN(tsv_mytext);
        schema_placeholder = ",".join(["?"] * len(schema_names))
        # PostgreSQL INFORMATION_SCHEMA.COLUMNS:
        # - https://www.postgresql.org/docs/9.1/static/infoschema-columns.html
        sql = translate_sql_qmark_to_percent("""
SELECT
    '' AS table_catalog,
    d.table_schema,
    d.table_name,
    d.column_name,
    d.is_nullable,
    d.column_type,
    d.column_comment,
    CASE WHEN COUNT(d.indrelid) > 0 THEN 1 ELSE 0 END AS indexed,
    MAX(d.indexed_fulltext) AS indexed_fulltext
FROM (
    SELECT
        -- c.table_catalog,  -- will always be the connection's database name
        c.table_schema,
        c.table_name,
        c.column_name,
        a.attnum as column_seq_num,
        c.is_nullable,
        pg_catalog.format_type(a.atttypid, a.atttypmod) as column_type,
        pgd.description AS column_comment,
        i.indrelid,
        CASE
            WHEN pg_get_indexdef(indexrelid) ~ 'USING (gin |gist )' THEN 1
            ELSE 0
        END AS indexed_fulltext
    FROM pg_catalog.pg_statio_all_tables AS t
    INNER JOIN information_schema.columns c ON (
        c.table_schema = t.schemaname
        AND c.table_name = t.relname
    )
    INNER JOIN pg_catalog.pg_attribute a ON (  -- one row per column
        a.attrelid = t.relid
        AND a.attname = c.column_name
    )
    LEFT JOIN pg_catalog.pg_index AS i ON (
        i.indrelid = t.relid  -- match on table
        AND i.indkey[0] = a.attnum  -- match on column sequence number
        AND i.indnatts = 1  -- one column in the index
    )
    LEFT JOIN pg_catalog.pg_description pgd ON (
        pgd.objoid = t.relid
        AND pgd.objsubid = c.ordinal_position
    )
    WHERE t.schemaname IN ({schema_placeholder})
) AS d
GROUP BY
    table_catalog,
    table_schema,
    table_name,
    column_name,
    is_nullable,
    column_type,
    column_comment
ORDER BY
    table_catalog,
    table_schema,
    table_name,
    column_name
        """.format(schema_placeholder=schema_placeholder))
        args = schema_names
        return sql, args

    @classmethod
    def _get_infodictlist_for_db(
            cls, connection: BaseDatabaseWrapper, vendor: str, db_name: str,
            schema_name: str) -> List[Dict[str, Any]]:
        log.debug("Fetching/caching database structure (for database {}, "
                  "schema {})...".format(db_name, schema_name))
        if db_name is None:
            raise ValueError("Use '', not None, for a blank db_name")
        if schema_name is None:
            raise ValueError("Use '', not None, for a blank schema_name")
        if vendor == VENDOR_MICROSOFT:
            if not db_name:
                raise ValueError("No db_name specified; required for MSSQL")
            if not schema_name:
                raise ValueError("No schema_name specified; required for MSSQL")  # noqa
            sql, args = cls._get_info_microsoft(db_name, [schema_name])
        elif vendor == VENDOR_POSTGRESQL:
            if db_name:
                raise ValueError("db_name specified; must be '' for PostgreSQL")  # noqa
            if not schema_name:
                raise ValueError("No schema_name specified; required for PostgreSQL")  # noqa
            sql, args = cls._get_info_postgres([schema_name])
        elif vendor == VENDOR_MYSQL:
            if db_name:
                raise ValueError("db_name specified; must be '' for MySQL")
            if not schema_name:
                raise ValueError("No schema_name specified; required for MySQL")  # noqa
            sql, args = cls._get_info_mysql(db_and_schema_name=schema_name)
        else:
            raise ValueError(
                "Don't know how to get metadata for "
                "connection.vendor=='{}'".format(vendor))
        # We execute this one directly, rather than using the Query class,
        # since this is a system rather than a per-user query.
        cursor = connection.cursor()
        # log.debug("sql = {}, args = {}".format(sql, repr(args)))
        cursor.execute(sql, args)
        results = dictfetchall(cursor)  # list of OrderedDicts
        # log.debug("results = {}".format(repr(results)))
        log.debug("... done")
        return results
        # Re passing multiple values to SQL via args:
        # - Don't circumvent the parameter protection against SQL injection.
        # - Too much hassle to use Django's ORM model here, though that would
        #   also be possible.
        # - http://stackoverflow.com/questions/907806
        # - Similarly via SQLAlchemy reflection/inspection.

    @django_cache_function(timeout=None)
    # @lru_cache(maxsize=None)
    def get_infodictlist(self) -> List[Dict[str, Any]]:
        connection = self._connection()
        vendor = connection.vendor
        results = []
        for dbinfo in settings.RESEARCH_DB_INFO:
            db_name = dbinfo['database']
            schema_name = dbinfo['schema']
            results.extend(self._get_infodictlist_for_db(connection, vendor,
                                                         db_name, schema_name))
        if not results:
            log.warning("ResearchDatabaseInfo._get_infodictlist(): no results "
                        "for 'research' database - misconfigured?")
        return results

    @django_cache_function(timeout=None)
    # @lru_cache(maxsize=None)
    def get_colinfolist(self) -> List[ColumnInfo]:
        infodictlist = self.get_infodictlist()
        return [ColumnInfo(**d) for d in infodictlist]

    @django_cache_function(timeout=None)
    # @lru_cache(maxsize=None)
    def get_colinfolist_by_tables(self) -> OrderedDict:
        colinfolist = self.get_colinfolist()
        table_to_colinfolist = {}
        for c in colinfolist:
            table_id = c.table_id()
            if table_id not in table_to_colinfolist:
                table_to_colinfolist[table_id] = []
            table_to_colinfolist[table_id].append(c)
        return OrderedDict(sorted(table_to_colinfolist.items()))

    @django_cache_function(timeout=None)
    # @lru_cache(maxsize=None)
    def get_colinfolist_by_schema(self) -> Dict[SchemaId, List[ColumnInfo]]:
        colinfolist = self.get_colinfolist()
        schema_to_colinfolist = {}  # type: Dict[SchemaId, List[ColumnInfo]]
        for c in colinfolist:
            table_id = c.table_id()
            schema = table_id.schema_id()
            if schema not in schema_to_colinfolist:
                schema_to_colinfolist[schema] = []
            schema_to_colinfolist[schema].append(c)
        return OrderedDict(sorted(schema_to_colinfolist.items()))

    def tables_containing_field(self,
                                fieldname: str) -> List[TableId]:
        """
        Returns a list of (db, schema, table) tuples.
        We won't use a SELECT on INFORMATION_SCHEMA here, since we already
        have the information.
        """
        columns = self.get_colinfolist()
        results = []
        for column in columns:
            if column.column_name == fieldname:
                table_id = column.table_id()
                if table_id not in results:
                    results.append(table_id)
        return results

    def text_columns(self, table_id: TableId,
                     min_length: int = 1) -> List[ColumnInfo]:
        results = []
        for column in self.get_colinfolist():
            if column.table_id() != table_id:
                continue
            if not is_sql_column_type_textual(column.column_type,
                                              min_length):
                continue
            results.append(column)
        return results

    @django_cache_function(timeout=None)
    # @lru_cache(maxsize=1000)
    def all_columns(self, table_id: TableId) -> List[ColumnInfo]:
        results = []
        for column in self.get_colinfolist():
            if column.table_id() != table_id:
                continue
            results.append(column)
        return results

    def get_tsv(self) -> str:
        return dictlist_to_tsv(self.get_infodictlist())

    def get_excel(self) -> bytes:
        wb = Workbook()
        wb.remove_sheet(wb.active)  # remove the autocreated blank sheet
        schema_colinfolist_dict = self.get_colinfolist_by_schema()
        for schema, colinfolist in schema_colinfolist_dict.items():
            ws = wb.create_sheet(title=schema.identifier(self.grammar))
            ws.append([
                "table_catalog", "table_schema", "table_name", "column_name",
                "is_nullable", "column_type", "column_comment",
                "indexed", "indexed_fulltext",
                "basetype", "full_identifier"
            ])
            for c in colinfolist:  # type: ColumnInfo
                ws.append([
                    c.table_catalog, c.table_schema, c.table_name,
                    c.column_name, c.is_nullable, c.column_type,
                    c.column_comment, c.indexed, c.indexed_fulltext,
                    c.basetype(), c.column_id().identifier(self.grammar),
                ])
        return excel_to_bytes(wb)

    @django_cache_function(timeout=None)
    # @lru_cache(maxsize=None)
    def get_tables(self) -> List[TableId]:
        tables = set()
        for column in self.get_colinfolist():
            tables.add(column.table_id())
        return sorted(list(tables))

    @django_cache_function(timeout=None)
    # @lru_cache(maxsize=1000)
    def table_contains_rid(self, table: TableId):
        target_rid_column = self.get_rid_column(table)
        for column in self.get_colinfolist():
            if column.column_id() == target_rid_column:
                return True
        return False

    def table_contains(self, table: TableId, column: ColumnId):
        for c in self.all_columns(table):
            if c.column_id() == column:
                return True
        return False

    @django_cache_function(timeout=None)
    # @lru_cache(maxsize=None)
    def get_mrid_linkable_patient_tables(self) -> List[TableId]:
        eligible_tables = set()
        for table in self.get_tables():
            schema = table.schema_id()
            if not self.does_db_schema_have_mrid(schema):
                continue
            if self.table_contains_rid(table):
                eligible_tables.add(table)
        return sorted(list(eligible_tables))


research_database_info = ResearchDatabaseInfo()



#!/usr/bin/env python
# crate_anon/anonymise/altermethod.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import datetime
import html
import logging
import os
import traceback
from typing import Any, Dict, List, Optional, Tuple, TYPE_CHECKING

from cardinal_pythonlib.rnc_datetime import (
    coerce_to_date,
    truncate_date_to_first_of_month,
)
from cardinal_pythonlib.rnc_extract_text import document_to_text
import regex

# don't import config: circular dependency would have to be sorted out
from crate_anon.anonymise.constants import ALTERMETHOD

if TYPE_CHECKING:
    from crate_anon.anonymise.config import Config
    from crate_anon.anonymise.ddr import DataDictionaryRow
    from crate_anon.anonymise.patient import Patient
    from crate_anon.common.hash import GenericHasher

log = logging.getLogger(__name__)


# =============================================================================
# Constants
# =============================================================================

HTML_TAG_RE = regex.compile('<[^>]*>')


# =============================================================================
# AlterMethod
# =============================================================================

class AlterMethod(object):
    def __init__(self,
                 config: "Config",
                 text_value: str = None,
                 scrub: bool = False,
                 truncate_date: bool = False,
                 extract_from_filename: bool = False,
                 extract_from_file_format: bool = False,  # new in v0.18.18
                 file_format_str: str = "",  # new in v0.18.18
                 extract_from_blob: bool = False,
                 skip_if_text_extract_fails: bool = False,
                 extract_ext_field: str = "",
                 hash_: bool = False,
                 hash_config_section: str = "",
                 # html_escape: bool = False,
                 html_unescape: bool = False,
                 html_untag: bool = False) -> None:
        self.config = config
        self.scrub = scrub
        self.truncate_date = truncate_date
        self.extract_from_blob = extract_from_blob
        self.extract_from_filename = extract_from_filename
        self.extract_from_file_format = extract_from_file_format
        self.file_format_str = file_format_str
        self.skip_if_text_extract_fails = skip_if_text_extract_fails
        self.extract_ext_field = extract_ext_field
        self.hash = hash_
        self.hash_config_section = hash_config_section
        self.hasher = None  # type: GenericHasher
        # self.html_escape = html_escape
        self.html_unescape = html_unescape
        self.html_untag = html_untag

        self.extract_text = (extract_from_filename or
                             extract_from_file_format or
                             extract_from_blob)

        if text_value is not None:
            self.set_from_text(text_value)
        if hash_:
            self.hasher = self.config.get_extra_hasher(
                self.hash_config_section)

    def set_from_text(self, value: str) -> None:
        """
        Convert the alter_method field (from the data dictionary) to a bunch of
        boolean/simple fields.
        """
        self.scrub = False
        self.truncate_date = False
        self.extract_text = False
        self.extract_from_blob = False
        self.extract_from_file_format = False
        self.file_format_str = ""
        self.extract_from_filename = False
        self.skip_if_text_extract_fails = False
        self.extract_ext_field = ""
        self.hash = False
        self.hash_config_section = ""

        def get_second_part(missing_description: str) -> str:
            if "=" not in value:
                raise ValueError(
                    "Bad format for alter method: {}".format(value))
            secondhalf = value[value.index("=") + 1:]
            if not secondhalf:
                raise ValueError(
                    "Missing {} in alter method: {}".format(
                        missing_description, value))
            return secondhalf

        if value == ALTERMETHOD.TRUNCATEDATE.value:
            self.truncate_date = True
        elif value == ALTERMETHOD.SCRUBIN.value:
            self.scrub = True
        elif value.startswith(ALTERMETHOD.BINARY_TO_TEXT.value):
            self.extract_text = True
            self.extract_from_blob = True
            self.extract_ext_field = get_second_part(
                "filename/extension field")
        elif value.startswith(ALTERMETHOD.FILENAME_FORMAT_TO_TEXT.value):
            self.extract_text = True
            self.extract_from_file_format = True
            self.file_format_str = get_second_part(
                "filename format field")
        elif value == ALTERMETHOD.FILENAME_TO_TEXT.value:
            self.extract_text = True
            self.extract_from_filename = True
        elif value == ALTERMETHOD.SKIP_IF_TEXT_EXTRACT_FAILS.value:
            self.skip_if_text_extract_fails = True
        elif value.startswith(ALTERMETHOD.HASH.value):
            self.hash = True
            self.hash_config_section = get_second_part("hash config section")
            self.hasher = self.config.get_extra_hasher(
                self.hash_config_section)
        # elif value == ALTERMETHOD.HTML_ESCAPE:
        #     self.html_escape = True
        elif value == ALTERMETHOD.HTML_UNESCAPE.value:
            self.html_unescape = True
        elif value == ALTERMETHOD.HTML_UNTAG.value:
            self.html_untag = True
        else:
            raise ValueError("Bad alter_method part: {}".format(value))

    def get_text(self) -> str:
        """
        Return the alter_method fragment from the working fields.
        """
        def two_part(altermethod: str, parameter: str):
            return altermethod + "=" + parameter

        if self.truncate_date:
            return ALTERMETHOD.TRUNCATEDATE.value
        if self.scrub:
            return ALTERMETHOD.SCRUBIN.value
        if self.extract_text:
            if self.extract_from_blob:
                return two_part(ALTERMETHOD.BINARY_TO_TEXT.value,
                                self.extract_ext_field)
            elif self.extract_from_file_format:
                return two_part(ALTERMETHOD.FILENAME_FORMAT_TO_TEXT.value,
                                self.file_format_str)
            else:  # plain filename
                return ALTERMETHOD.FILENAME_TO_TEXT.value
        if self.skip_if_text_extract_fails:
            return ALTERMETHOD.SKIP_IF_TEXT_EXTRACT_FAILS.value
        if self.hash:
            return two_part(ALTERMETHOD.HASH.value,
                            self.hash_config_section)
        # if self.html_escape:
        #     return ALTERMETHOD.HTML_ESCAPE.value
        if self.html_unescape:
            return ALTERMETHOD.HTML_UNESCAPE.value
        if self.html_untag:
            return ALTERMETHOD.HTML_UNTAG.value
        return ""

    def alter(self,
              value: Any,  # value of interest
              ddr: "DataDictionaryRow",  # corresponding DataDictionaryRow
              row: List[Any],  # all values in row
              ddrows: List["DataDictionaryRow"],  # all of them
              patient: "Patient" = None) -> Tuple[Any, bool]:
        """
        Performs the alteration.
        Returns:
            Tuple[newvalue: Any, skiprow: bool]
        """

        if self.scrub:
            return self._scrub_func(value, patient), False

        if self.truncate_date:
            return self._truncate_date_func(value), False

        if self.extract_text:
            value, extracted = self._extract_text_func(value, row, ddrows)
            if not extracted and ddr.skip_row_if_extract_text_fails():
                log.debug("Skipping row as text extraction failed")
                return None, True
            return value, False

        if self.hash:
            assert self.hasher is not None
            return self.hasher.hash(value), False

        # if alter_method.html_escape:
        #     return html.escape(value), False

        if self.html_unescape:
            return html.unescape(value), False

        if self.html_untag:
            return self._html_untag_func(value), False

        if self.skip_if_text_extract_fails:
            # Modifies other alter methods; doesn't do anything itself
            return value, True

    @staticmethod
    def _scrub_func(value: Any, patient: "Patient") -> Any:
        if value is None:
            return None
        # Main point of anonymisation!
        return patient.scrub(value)

    @staticmethod
    def _truncate_date_func(value: Any) -> Optional[datetime.datetime]:
        try:
            value = coerce_to_date(value)
            return truncate_date_to_first_of_month(value)
        except (ValueError, OverflowError):
            log.warning(
                "Invalid date received to "
                "{ALTERMETHOD.TRUNCATEDATE} method: {v}".format(
                    ALTERMETHOD=ALTERMETHOD, v=value))
            return None

    @staticmethod
    def _html_untag_func(text: str) -> str:
        # Lots of ways...
        # -- xml.etree, for well-formed XML
        #    http://stackoverflow.com/questions/9662346
        # return ''.join(xml.etree.ElementTree.fromstring(text).itertext())
        # -- html.parser
        #    http://stackoverflow.com/questions/753052
        # -- lxml (but needs source build on Windows):
        #    http://www.neuraladvance.com/removing-html-from-python-strings.html
        #    http://lxml.de/
        # -- regex/re
        #    http://stackoverflow.com/questions/3662142
        return HTML_TAG_RE.sub('', text)

    def _extract_text_func(
            self, value: Any, row: List[Any],
            ddrows: List["DataDictionaryRow"]) -> Tuple[Optional[str], bool]:
        """
        Take a field's value and return extracted text, for file-related 
        fields, where the DD row indicates that this field contains a filename 
        or a BLOB.
    
        Returns tuple: value, extracted
        """
        filename = None
        blob = None

        # Work out either a full filename, or a BLOB.
        if self.extract_from_filename:
            # The database contains a plain and full filename.
            filename = value
            _, extension = os.path.splitext(filename)
            log.info("extract_text: disk file, filename={}".format(filename))

        elif self.extract_from_file_format:
            # The database contains a filename. However, it may not be a full
            # path. For example, in RiO, we have fields like
            #   dbo.ClientDocument.Path, e.g. '1-1-20121023-1000001-LET.pdf'
            #   dbo.ClientDocument.ClientID, e.g. '1000001-LET.pdf'
            # and the disk file might be
            #   C:\some_base_directory\1000001\Docs\1-1-20121023-1000001-LET.pdf
            # We could specify this as a file spec:
            #   "C:\some_base_directory\{ClientID}\{Path}".
            # In principle, this might need to be field-specific, so it could
            # go in the data dictionary (rather than as a setting that's
            # constant across an entire anonymisation run).
            # Let's introduce ALTERMETHOD.FILENAME_FORMAT_TO_TEXT, in v0.18.18.
            #
            # Create a dictionary of column name -> value
            ffdict = {}  # type: Dict[str, Any]
            for i, ddr in enumerate(ddrows):
                ffdict[ddr.src_field] = row[i]
            # Use that dictionary with the format string to make the filename
            log.debug("extract_text: file_format_str={}, ffdict={}".format(
                repr(self.file_format_str), repr(ffdict)))
            filename = self.file_format_str.format(**ffdict)
            _, extension = os.path.splitext(filename)
            log.info("extract_text: disk file, filename={}".format(filename))

        else:
            # The database contains the BLOB itself. However, we'd also like to
            # know the file type, here from its extension. We look for another
            # field that contains the extension, marked as such using
            # alter_method.extract_ext_field in the data dictionary.
            blob = value
            extindex = next(
                (i for i, ddr in enumerate(ddrows)
                    if ddr.src_field == self.extract_ext_field),
                None)
            if extindex is None:
                # Configuration error
                raise ValueError(
                    "Bug: missing extension field for "
                    "alter_method={}".format(self.get_text()))
            extension = row[extindex]
            log.info("extract_text: database BLOB, extension={}".format(
                extension))

        # Is it a permissible filename?
        if not self.config.extract_text_extension_permissible(extension):
            log.info("Extension {} not permissible; skipping".format(
                repr(extension)))
            return None, False

        # Extract text from the file (given its filename), or from a BLOB.
        try:
            value = document_to_text(filename=filename,
                                     blob=blob,
                                     extension=extension,
                                     plain=self.config.extract_text_plain,
                                     width=self.config.extract_text_width)
        except Exception as e:
            # Runtime error
            traceback.print_exc()  # full details, please
            log.error("Caught exception from document_to_text: {}".format(e))
            return None, False
        return value, True



# -*- coding: utf-8 -*-
# Generated by Django 1.9.8 on 2017-02-06 16:17
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import picklefield.fields


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('research', '0002_auto_20170203_1348'),
    ]

    operations = [
        migrations.CreateModel(
            name='PatientExplorer',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False)),
                ('patient_multiquery', picklefield.fields.PickledObjectField(editable=False, null=True, verbose_name='Pickled PatientMultiQuery')),  # noqa
                ('active', models.BooleanField(default=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('deleted', models.BooleanField(default=False, verbose_name="Deleted from the user's perspective. Audited queries are never properly deleted.")),  # noqa
                ('audited', models.BooleanField(default=False)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),  # noqa
            ],
        ),
        migrations.CreateModel(
            name='PatientExplorerAudit',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False)),
                ('when', models.DateTimeField(auto_now_add=True)),
                ('count_only', models.BooleanField(default=False)),
                ('n_records', models.IntegerField(default=0)),
                ('failed', models.BooleanField(default=False)),
                ('fail_msg', models.TextField()),
                ('patient_explorer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='research.PatientExplorer')),  # noqa
            ],
        ),
    ]



##!/usr/bin/env python
# crate_anon/common/parallel.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import logging
from typing import Any

from crate_anon.common.hash import hash64

log = logging.getLogger(__name__)


def is_my_job_by_int(value: int, tasknum: int, ntasks: int) -> bool:
    if ntasks == 1:
        return True
    return value % ntasks == tasknum


def is_my_job_by_hash(value: Any, tasknum: int, ntasks: int) -> bool:
    """
    We convert some non-integer thing into a deterministic but roughly
    randomly distributed integer using hash64. That produces a signed integer,
    which is OK because % works nonetheless.

    We use this function to parallelize for non-integer PKs.

    This is less efficient than dividing the work up via SQL, because we have
    to fetch/hash something.

    Perform this test ASAP in loops, for speed.
    """
    if ntasks == 1:
        return True
    return hash64(value) % ntasks == tasknum



#!/usr/bin/env python
# crate_anon/anonymise/config.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================

Config class for CRATE anonymiser.

Thoughts on configuration method

-   First version used a Config() class, which initializes with blank values.
    The anonymise_cli.py file creates a config singleton and passes it around.
    Then when its set() method is called, it reads a config file and
    instantiates its settings.
    An option exists to print a draft config without ever reading one from
    disk.

    Advantage: easy to start the program without a valid config file (e.g. to
        print one).
    Disadvantage: modules can't be sure that a config is properly instantiated
        before they are loaded, so you can't easily define a class according to
        config settings (you'd have to have a class factory, which gets ugly).

-   The Django method is to have a configuration file (e.g. settings.py, which
    can import from other things) that is read by Django and then becomes
    importable by anything at startup as "django.conf.settings". (I've added
    local settings via an environment variable.) The way Django knows where
    to look is via this in manage.py:

        os.environ.setdefault("DJANGO_SETTINGS_MODULE",
                              "crate_anon.crateweb.config.settings")

    Advantage: setting the config file via an environment variable (read when
        the config file loads) allows guaranteed config existence as other
        modules start.
    Further advantage: config filenames not on command line, therefore not
        visible to ps.
    Disadvantage: how do you override with a command-line (argparse) setting?
        ... though: who cares?
    To print a config using that file: raise an exception on nonexistent
        config, and catch it with a special entry point script.

-   See also
    http://stackoverflow.com/questions/7443366/argument-passing-strategy-environment-variables-vs-command-line  # noqa
"""

# =============================================================================
# Imports
# =============================================================================

import codecs
import fnmatch
import logging
import os
import sys
from typing import Any, Dict, List, Optional, TYPE_CHECKING

import regex
from cardinal_pythonlib.rnc_db import (
    ensure_valid_field_name,
    ensure_valid_table_name,
)
from cardinal_pythonlib.rnc_log import remove_all_logger_handlers
from sqlalchemy import BigInteger, create_engine, String
from sqlalchemy.dialects.mssql.base import dialect as mssql_dialect
from sqlalchemy.dialects.mysql.base import dialect as mysql_dialect
from sqlalchemy.engine import Engine
from sqlalchemy.sql.sqltypes import TypeEngine

from crate_anon.anonymise.constants import (
    CONFIG_ENV_VAR,
    DEFAULT_CHUNKSIZE,
    DEFAULT_REPORT_EVERY,
    DEFAULT_MAX_ROWS_BEFORE_COMMIT,
    DEFAULT_MAX_BYTES_BEFORE_COMMIT,
    SEP,
)
from crate_anon.anonymise.dd import DataDictionary
from crate_anon.anonymise.scrub import (
    NonspecificScrubber,
    WordList,
)
from crate_anon.common.extendedconfigparser import ExtendedConfigParser
from crate_anon.common.formatting import sizeof_fmt
from crate_anon.common.hash import GenericHasher, make_hasher
from crate_anon.common.sql import TransactionSizeLimiter
from crate_anon.common.sqla import (
    hack_in_mssql_xml_type,
    is_sqlatype_integer,
    monkeypatch_TableClause,
)

if TYPE_CHECKING:
    from crate_anon.anonymise.dbholder import DatabaseHolder

log = logging.getLogger(__name__)

# The Config class is loaded very early, via the nasty singleton.
# This is therefore the appropriate moment to make any changes to SQLAlchemy.
monkeypatch_TableClause()
hack_in_mssql_xml_type()  # support XML type under SQL Server


# =============================================================================
# Config/databases
# =============================================================================

class DatabaseSafeConfig(object):
    """Class representing non-sensitive configuration information about a
    source database."""

    def __init__(self, parser: ExtendedConfigParser, section: str) -> None:
        """Read from a configparser section."""
        if not parser.has_section(section):
            raise ValueError("config missing section: " + section)

        def opt_str(option: str) -> str:
            return parser.get(section, option, fallback=None)

        def opt_multiline(option: str, as_words: bool = True) -> List[str]:
            return parser.get_str_list(section, option, as_words=as_words)

        def opt_bool(option: str, default: bool) -> bool:
            return parser.getboolean(section, option, fallback=default)

        def opt_int(option: str, default: Optional[int]) -> Optional[int]:
            return parser.get_int_default_if_failure(section, option, default)

        def opt_multiline_csv_pairs(option: str) -> Dict[str, str]:
            d = {}
            lines = opt_multiline(option, as_words=False)
            for line in lines:
                pair = [item.strip() for item in line.split(",")]
                if len(pair) != 2:
                    raise ValueError("For option {}: specify items as a list "
                                     "of comma-separated pairs".format(option))
                d[pair[0]] = pair[1]
            return d

        self.ddgen_omit_by_default = opt_bool(
            'ddgen_omit_by_default', True)
        self.ddgen_omit_fields = opt_multiline('ddgen_omit_fields')
        self.ddgen_include_fields = opt_multiline('ddgen_include_fields')

        self.ddgen_allow_no_patient_info = opt_bool(
            'ddgen_allow_no_patient_info', False)
        self.ddgen_per_table_pid_field = opt_str('ddgen_per_table_pid_field')
        self.ddgen_add_per_table_pids_to_scrubber = opt_bool(
            'ddgen_add_per_table_pids_to_scrubber', False)
        self.ddgen_master_pid_fieldname = opt_str('ddgen_master_pid_fieldname')
        self.ddgen_table_blacklist = opt_multiline('ddgen_table_blacklist')
        self.ddgen_table_whitelist = opt_multiline('ddgen_table_whitelist')
        self.ddgen_table_require_field_absolute = opt_multiline(
            'ddgen_table_require_field_absolute')
        self.ddgen_table_require_field_conditional = opt_multiline_csv_pairs(
            'ddgen_table_require_field_conditional')
        self.ddgen_field_blacklist = opt_multiline('ddgen_field_blacklist')
        self.ddgen_field_whitelist = opt_multiline('ddgen_field_whitelist')
        self.ddgen_pk_fields = opt_multiline('ddgen_pk_fields')

        self.ddgen_constant_content = opt_bool(
            'ddgen_constant_content', False)
        self.ddgen_constant_content_tables = opt_str(
            'ddgen_constant_content_tables')
        self.ddgen_nonconstant_content_tables = opt_str(
            'ddgen_nonconstant_content_tables')
        self.ddgen_addition_only = opt_bool('ddgen_addition_only', False)
        self.ddgen_addition_only_tables = opt_str('ddgen_addition_only_tables')
        self.ddgen_deletion_possible_tables = opt_str(
            'ddgen_deletion_possible_tables')

        self.ddgen_pid_defining_fieldnames = opt_multiline(
            'ddgen_pid_defining_fieldnames')
        self.ddgen_scrubsrc_patient_fields = opt_multiline(
            'ddgen_scrubsrc_patient_fields')
        self.ddgen_scrubsrc_thirdparty_fields = opt_multiline(
            'ddgen_scrubsrc_thirdparty_fields')
        self.ddgen_scrubsrc_thirdparty_xref_pid_fields = opt_multiline(
            'ddgen_scrubsrc_thirdparty_xref_pid_fields')
        self.ddgen_required_scrubsrc_fields = opt_multiline(
            'ddgen_required_scrubsrc_fields')
        self.ddgen_scrubmethod_code_fields = opt_multiline(
            'ddgen_scrubmethod_code_fields')
        self.ddgen_scrubmethod_date_fields = opt_multiline(
            'ddgen_scrubmethod_date_fields')
        self.ddgen_scrubmethod_number_fields = opt_multiline(
            'ddgen_scrubmethod_number_fields')
        self.ddgen_scrubmethod_phrase_fields = opt_multiline(
            'ddgen_scrubmethod_phrase_fields')
        self.ddgen_safe_fields_exempt_from_scrubbing = opt_multiline(
            'ddgen_safe_fields_exempt_from_scrubbing')
        self.ddgen_min_length_for_scrubbing = opt_int(
            'ddgen_min_length_for_scrubbing', 0)

        self.ddgen_truncate_date_fields = opt_multiline(
            'ddgen_truncate_date_fields')
        self.ddgen_filename_to_text_fields = opt_multiline(
            'ddgen_filename_to_text_fields')

        self.bin2text_dict = opt_multiline_csv_pairs(
            'ddgen_binary_to_text_field_pairs')
        self.ddgen_skip_row_if_extract_text_fails_fields = opt_multiline(
            'ddgen_skip_row_if_extract_text_fails_fields')
        self.ddgen_rename_tables_remove_suffixes = opt_multiline(
            'ddgen_rename_tables_remove_suffixes', as_words=True)

        self.ddgen_index_fields = opt_multiline('ddgen_index_fields')
        self.ddgen_allow_fulltext_indexing = opt_bool(
            'ddgen_allow_fulltext_indexing', True)

        self.ddgen_force_lower_case = opt_bool('ddgen_force_lower_case', True)
        self.ddgen_convert_odd_chars_to_underscore = opt_bool(
            'ddgen_convert_odd_chars_to_underscore', True)

        self.debug_row_limit = opt_int('debug_row_limit', 0)
        self.debug_limited_tables = opt_multiline('debug_limited_tables')

        self.ddgen_patient_opt_out_fields = opt_multiline(
            'ddgen_patient_opt_out_fields')

        self.ddgen_extra_hash_fields = opt_multiline_csv_pairs(
            'ddgen_extra_hash_fields')
        # ... key: fieldspec
        # ... value: hash_config_section_name

        self.PidType = BigInteger
        self.MpidType = BigInteger

    def is_table_blacklisted(self, table: str) -> bool:
        for white in self.ddgen_table_whitelist:
            r = regex.compile(fnmatch.translate(white), regex.IGNORECASE)
            if r.match(table):
                return False
        for black in self.ddgen_table_blacklist:
            r = regex.compile(fnmatch.translate(black), regex.IGNORECASE)
            if r.match(table):
                return True
        return False

    def is_field_blacklisted(self, field: str) -> bool:
        for white in self.ddgen_field_whitelist:
            r = regex.compile(fnmatch.translate(white), regex.IGNORECASE)
            if r.match(field):
                return True
        for black in self.ddgen_field_blacklist:
            r = regex.compile(fnmatch.translate(black), regex.IGNORECASE)
            if r.match(field):
                return True
        return False

    def does_table_fail_minimum_fields(self, colnames: List[str]) -> bool:
        for abs_req in self.ddgen_table_require_field_absolute:
            if abs_req not in colnames:
                return True
        for if_field, then_field in self.ddgen_table_require_field_conditional.items():  # noqa
            if if_field in colnames and then_field not in colnames:
                return True
        return False


# =============================================================================
# ExtraHashconfig
# =============================================================================

def get_extra_hasher(parser: ExtendedConfigParser,
                     section: str) -> GenericHasher:
    """Read from a configparser section."""
    if not parser.has_section(section):
        raise ValueError("config missing section: " + section)

    def opt_str(option: str) -> str:
        return parser.get(section, option, fallback=None)

    hash_method = opt_str("hash_method")
    secret_key = opt_str("secret_key")
    return make_hasher(hash_method, secret_key)


# =============================================================================
# Config
# =============================================================================

class Config(object):
    """Class representing the main configuration."""

    def __init__(self, open_databases: bool = True) -> None:
        """
        Read config from file
        """
        # Get filename
        try:
            self.config_filename = os.environ[CONFIG_ENV_VAR]
            assert self.config_filename
        except (KeyError, AssertionError):
            print(
                "You must set the {} environment variable to point to a CRATE "
                "anonymisation config file, or specify it on the command "
                "line. Run crate_print_demo_anon_config "
                "to see a specimen config.".format(CONFIG_ENV_VAR))
            sys.exit(1)

        # Read config from file.
        parser = ExtendedConfigParser()
        log.info("Reading config file: {}".format(self.config_filename))
        parser.read_file(codecs.open(self.config_filename, "r", "utf8"))
        section = "main"

        def opt_str(option: str) -> str:
            return parser.get(section, option, fallback=None)

        def opt_multiline(option: str) -> List[str]:
            return parser.get_str_list(section, option)

        def opt_multiline_int(option: str,
                              minimum: int = None,
                              maximum: int = None) -> List[int]:
            return parser.get_int_list(section, option, minimum=minimum,
                                       maximum=maximum, suppress_errors=False)

        def opt_bool(option: str, default: bool) -> bool:
            return parser.getboolean(section, option, fallback=default)

        def opt_int(option: str, default: Optional[int]) -> Optional[int]:
            return parser.get_int_default_if_failure(section, option, default)

        def opt_pyvalue_list(option: str, default: Any = None) -> Any:
            return parser.get_pyvalue_list(section, option, default=default)

        def get_database(section_: str,
                         name: str,
                         srccfg_: DatabaseSafeConfig = None,
                         with_session: bool = False,
                         with_conn: bool = True,
                         reflect: bool = True) -> "DatabaseHolder":
            return parser.get_database(section_,
                                       dbname=name,
                                       srccfg=srccfg_,
                                       with_session=with_session,
                                       with_conn=with_conn,
                                       reflect=reflect)

        def get_sqlatype(sqlatype: str, default: TypeEngine) -> TypeEngine:
            if not sqlatype:
                return default
            if sqlatype == "BigInteger":
                return BigInteger
            r = regex.compile(r"String\((\d+)\)")  # e.g. String(50)
            try:
                m = r.match(sqlatype)
                length = int(m.group(1))
                return String(length)
            except (AttributeError, ValueError):
                raise ValueError("Bad SQLAlchemy type specification for "
                                 "PID/MPID columns: {}".format(repr(sqlatype)))

        # ---------------------------------------------------------------------
        # Data dictionary
        # ---------------------------------------------------------------------

        self.data_dictionary_filename = opt_str('data_dictionary_filename')

        # ---------------------------------------------------------------------
        # Critical field types
        # ---------------------------------------------------------------------

        self.PidType = get_sqlatype(opt_str('sqlatype_pid'), BigInteger)
        self.pidtype_is_integer = is_sqlatype_integer(self.PidType)
        self.MpidType = get_sqlatype(opt_str('sqlatype_mpid'), BigInteger)
        self.mpidtype_is_integer = is_sqlatype_integer(self.MpidType)

        # ---------------------------------------------------------------------
        # Encryption phrases/passwords
        # ---------------------------------------------------------------------

        self.hash_method = opt_str('hash_method')
        self.per_table_patient_id_encryption_phrase = opt_str(
            'per_table_patient_id_encryption_phrase')
        self.master_patient_id_encryption_phrase = opt_str(
            'master_patient_id_encryption_phrase')
        self.change_detection_encryption_phrase = opt_str(
            'change_detection_encryption_phrase')
        _extra_hash_config_section_names = opt_multiline(
            "extra_hash_config_sections")

        self.extra_hashers = {}  # type: Dict[str, GenericHasher]
        for hasher_name in _extra_hash_config_section_names:
            self.extra_hashers[hasher_name] = get_extra_hasher(parser,
                                                               hasher_name)
        # Load encryption keys and create hashers
        dummyhash = make_hasher(self.hash_method, "dummysalt")
        encrypted_length = dummyhash.output_length()

        self.SqlTypeEncryptedPid = String(encrypted_length)
        self.sqltype_encrypted_pid_as_sql = str(self.SqlTypeEncryptedPid)
        # ... VARCHAR(32) for MD5; VARCHAR(64) for SHA-256; VARCHAR(128) for
        # SHA-512.

        if not self.per_table_patient_id_encryption_phrase:
            raise ValueError("Missing per_table_patient_id_encryption_phrase")
        self.primary_pid_hasher = make_hasher(
            self.hash_method, self.per_table_patient_id_encryption_phrase)

        if not self.master_patient_id_encryption_phrase:
            raise ValueError("Missing master_patient_id_encryption_phrase")
        self.master_pid_hasher = make_hasher(
            self.hash_method, self.master_patient_id_encryption_phrase)

        if not self.change_detection_encryption_phrase:
            raise ValueError("Missing change_detection_encryption_phrase")
        self.change_detection_hasher = make_hasher(
            self.hash_method, self.change_detection_encryption_phrase)

        # ---------------------------------------------------------------------
        # Text extraction
        # ---------------------------------------------------------------------

        self.extract_text_extensions_case_sensitive = opt_bool(
            'extract_text_extensions_case_sensitive', False)
        self.extract_text_extensions_permitted = opt_multiline(
            'extract_text_extensions_permitted')
        self.extract_text_extensions_prohibited = opt_multiline(
            'extract_text_extensions_prohibited')
        self.extract_text_plain = opt_bool('extract_text_plain', False)
        self.extract_text_width = opt_int('extract_text_width', 80)

        # ---------------------------------------------------------------------
        # Anonymisation
        # ---------------------------------------------------------------------

        self.replace_patient_info_with = opt_str('replace_patient_info_with')
        self.replace_third_party_info_with = opt_str(
            'replace_third_party_info_with')
        self.replace_nonspecific_info_with = opt_str(
            'replace_nonspecific_info_with')
        self.thirdparty_xref_max_depth = opt_int('thirdparty_xref_max_depth',
                                                 1)
        self.string_max_regex_errors = opt_int('string_max_regex_errors', 0)
        self.min_string_length_for_errors = opt_int(
            'min_string_length_for_errors', 1)
        self.min_string_length_to_scrub_with = opt_int(
            'min_string_length_to_scrub_with', 2)
        self.scrub_all_uk_postcodes = opt_bool('scrub_all_uk_postcodes', False)
        self.anonymise_codes_at_word_boundaries_only = opt_bool(
            'anonymise_codes_at_word_boundaries_only', True)
        self.anonymise_dates_at_word_boundaries_only = opt_bool(
            'anonymise_dates_at_word_boundaries_only', True)
        self.anonymise_numbers_at_word_boundaries_only = opt_bool(
            'anonymise_numbers_at_word_boundaries_only', False)
        self.anonymise_numbers_at_numeric_boundaries_only = opt_bool(
            'anonymise_numbers_at_numeric_boundaries_only', True)
        self.anonymise_strings_at_word_boundaries_only = opt_bool(
            'anonymise_strings_at_word_boundaries_only', True)

        self.scrub_string_suffixes = opt_multiline('scrub_string_suffixes')
        self.whitelist_filenames = opt_multiline('whitelist_filenames')
        self.blacklist_filenames = opt_multiline('blacklist_filenames')
        self.scrub_all_numbers_of_n_digits = opt_multiline_int(
            'scrub_all_numbers_of_n_digits', minimum=1)

        if not self.extract_text_extensions_case_sensitive:
            self.extract_text_extensions_permitted = [
                x.upper() for x in self.extract_text_extensions_permitted]
            self.extract_text_extensions_permitted = [
                x.upper() for x in self.extract_text_extensions_permitted]

        # Whitelist, blacklist, nonspecific scrubber
        self.whitelist = WordList(
            filenames=self.whitelist_filenames,
            hasher=self.change_detection_hasher,
        )
        self.blacklist = WordList(
            filenames=self.blacklist_filenames,
            replacement_text=self.replace_nonspecific_info_with,
            hasher=self.change_detection_hasher,
            at_word_boundaries_only=(
                self.anonymise_strings_at_word_boundaries_only),
            max_errors=0,
        )
        self.nonspecific_scrubber = NonspecificScrubber(
            replacement_text=self.replace_nonspecific_info_with,
            hasher=self.change_detection_hasher,
            anonymise_codes_at_word_boundaries_only=(
                self.anonymise_codes_at_word_boundaries_only),
            anonymise_numbers_at_word_boundaries_only=(
                self.anonymise_numbers_at_word_boundaries_only),
            blacklist=self.blacklist,
            scrub_all_numbers_of_n_digits=self.scrub_all_numbers_of_n_digits,
            scrub_all_uk_postcodes=self.scrub_all_uk_postcodes,
        )

        # ---------------------------------------------------------------------
        # Output fields and formatting
        # ---------------------------------------------------------------------

        self.mapping_patient_id_fieldname = opt_str(
            'mapping_patient_id_fieldname')
        self.research_id_fieldname = opt_str('research_id_fieldname')
        self.trid_fieldname = opt_str('trid_fieldname')
        self.mapping_master_id_fieldname = opt_str(
            'mapping_master_id_fieldname')
        self.master_research_id_fieldname = opt_str(
            'master_research_id_fieldname')
        self.source_hash_fieldname = opt_str('source_hash_fieldname')
        self.date_to_text_format = opt_str('date_to_text_format')
        self.datetime_to_text_format = opt_str('datetime_to_text_format')
        self.append_source_info_to_comment = opt_bool(
            'append_source_info_to_comment', True)

        # ---------------------------------------------------------------------
        # Destination database configuration
        # ---------------------------------------------------------------------

        self.max_rows_before_commit = opt_int('max_rows_before_commit',
                                              DEFAULT_MAX_ROWS_BEFORE_COMMIT)
        self.max_bytes_before_commit = opt_int('max_bytes_before_commit',
                                               DEFAULT_MAX_BYTES_BEFORE_COMMIT)
        self.temporary_tablename = opt_str('temporary_tablename')

        # ---------------------------------------------------------------------
        # Databases
        # ---------------------------------------------------------------------

        destination_database_cfg_section = opt_str('destination_database')
        self._destination_database_url = parser.get_str(
            destination_database_cfg_section, 'url', required=True)
        admin_database_cfg_section = opt_str('admin_database')
        if destination_database_cfg_section == admin_database_cfg_section:
            raise ValueError(
                "Destination and admin databases mustn't be the same")
        source_database_cfg_sections = opt_multiline('source_databases')
        self.source_db_names = source_database_cfg_sections
        if destination_database_cfg_section in source_database_cfg_sections:
            raise ValueError("Destination database mustn't be listed as a "
                             "source database")
        if admin_database_cfg_section in source_database_cfg_sections:
            raise ValueError("Admin database mustn't be listed as a "
                             "source database")

        self.destdb = get_database(destination_database_cfg_section,
                                   name=destination_database_cfg_section,
                                   with_session=open_databases,
                                   with_conn=False,
                                   reflect=False)
        if not self.destdb:
            raise ValueError("Destination database misconfigured")
        if open_databases:
            self.dest_dialect = self.destdb.engine.dialect
        else:  # in context of web framework, some sort of default
            self.dest_dialect = mysql_dialect
        self._destdb_transaction_limiter = TransactionSizeLimiter(
            session=self.destdb.session,
            max_bytes_before_commit=self.max_bytes_before_commit,
            max_rows_before_commit=self.max_rows_before_commit
        )

        self.admindb = get_database(admin_database_cfg_section,
                                    name=admin_database_cfg_section,
                                    with_session=open_databases,
                                    with_conn=False,
                                    reflect=open_databases)
        if not self.admindb:
            raise ValueError("Admin database misconfigured")

        self.sources = {}
        self.src_dialects = {}
        for sourcedb_name in source_database_cfg_sections:
            log.info("Adding source database: {}".format(sourcedb_name))
            srccfg = DatabaseSafeConfig(parser, sourcedb_name)
            srcdb = get_database(sourcedb_name,
                                 srccfg_=srccfg,
                                 name=sourcedb_name,
                                 with_session=open_databases,
                                 with_conn=False,
                                 reflect=open_databases)
            if not srcdb:
                raise ValueError("Source database {} misconfigured".format(
                    sourcedb_name))
            self.sources[sourcedb_name] = srcdb
            if open_databases:
                self.src_dialects[sourcedb_name] = srcdb.engine.dialect
            else:  # in context of web framework
                self.src_dialects[sourcedb_name] = mssql_dialect

        # ---------------------------------------------------------------------
        # Processing options
        # ---------------------------------------------------------------------

        self.debug_max_n_patients = opt_int('debug_max_n_patients', 0)
        self.debug_pid_list = opt_multiline('debug_pid_list')

        # ---------------------------------------------------------------------
        # Opting out entirely
        # ---------------------------------------------------------------------

        self.optout_pid_filenames = opt_multiline('optout_pid_filenames')
        self.optout_mpid_filenames = opt_multiline('optout_mpid_filenames')
        self.optout_col_values = opt_pyvalue_list('optout_col_values')

        # ---------------------------------------------------------------------
        # Rest of initialization
        # ---------------------------------------------------------------------

        self.dd = DataDictionary(self)

        self.rows_in_transaction = 0
        self.bytes_in_transaction = 0
        self.rows_inserted_per_table = {}
        self.warned_re_limits = {}

        self.report_every_n_rows = DEFAULT_REPORT_EVERY
        self.chunksize = DEFAULT_CHUNKSIZE
        self.debug_scrubbers = False
        self.save_scrubbers = False
        
        self._src_bytes_read = 0
        self._dest_bytes_written = 0
        self._echo = False

    def get_destdb_engine_outside_transaction(
            self, encoding: str = 'utf-8') -> Engine:
        url = self._destination_database_url
        return create_engine(
            url,
            encoding=encoding,
            echo=self._echo,
            connect_args={'autocommit': True})  # for pyodbc
        # https://github.com/mkleehammer/pyodbc/wiki/Database-Transaction-Management  # noqa

    def overall_progress(self) -> str:
        return "{} read, {} written".format(
            sizeof_fmt(self._src_bytes_read),
            sizeof_fmt(self._dest_bytes_written))

    def load_dd(self, check_against_source_db: bool = True) -> None:
        log.info(SEP + "Loading data dictionary: {}".format(
            self.data_dictionary_filename))
        self.dd.read_from_file(self.data_dictionary_filename)
        self.dd.check_valid(
            prohibited_fieldnames=[self.source_hash_fieldname,
                                   self.trid_fieldname],
            check_against_source_db=check_against_source_db)
        self.init_row_counts()

    def init_row_counts(self) -> None:
        """Initialize row counts for all source tables."""
        self.rows_inserted_per_table = {}
        for db_table_tuple in self.dd.get_src_db_tablepairs():
            self.rows_inserted_per_table[db_table_tuple] = 0
            self.warned_re_limits[db_table_tuple] = False

    def check_valid(self) -> None:
        """Raise exception if config is invalid."""

        # Destination databases
        if not self.destdb:
            raise ValueError("No destination database specified.")
        if not self.admindb:
            raise ValueError("No admin database specified.")

        # Test table names
        if not self.temporary_tablename:
            raise ValueError("No temporary_tablename specified.")
        ensure_valid_table_name(self.temporary_tablename)

        # Test field names
        def validate_fieldattr(name):
            if not getattr(self, name):
                raise ValueError("Blank fieldname: " + name)
            ensure_valid_field_name(getattr(self, name))

        specialfieldlist = [
            "mapping_patient_id_fieldname",
            "research_id_fieldname",
            "trid_fieldname",
            "master_research_id_fieldname",
            "mapping_master_id_fieldname",
            "source_hash_fieldname",
        ]
        fieldset = set()
        for attrname in specialfieldlist:
            validate_fieldattr(attrname)
            fieldset.add(getattr(self, attrname))
        if len(fieldset) != len(specialfieldlist):
            raise ValueError(
                "Config: these must all be DIFFERENT fieldnames: " +
                ",".join(specialfieldlist))

        # Test strings
        if not self.replace_patient_info_with:
            raise ValueError("Blank replace_patient_info_with")
        if not self.replace_third_party_info_with:
            raise ValueError("Blank replace_third_party_info_with")
        if not self.replace_nonspecific_info_with:
            raise ValueError("Blank replace_nonspecific_info_with")
        replacements = list({self.replace_patient_info_with,
                             self.replace_third_party_info_with,
                             self.replace_nonspecific_info_with})
        if len(replacements) != 3:
            raise ValueError(
                "Inadvisable: replace_patient_info_with, "
                "replace_third_party_info_with, and "
                "replace_nonspecific_info_with should all be distinct")

        # Regex
        if self.string_max_regex_errors < 0:
            raise ValueError("string_max_regex_errors < 0, nonsensical")
        if self.min_string_length_for_errors < 1:
            raise ValueError("min_string_length_for_errors < 1, nonsensical")
        if self.min_string_length_to_scrub_with < 1:
            raise ValueError(
                "min_string_length_to_scrub_with < 1, nonsensical")

        # Source databases
        if not self.sources:
            raise ValueError("No source databases specified.")
        for dbname, dbinfo in self.sources.items():
            cfg = dbinfo.srccfg
            if not cfg.ddgen_allow_no_patient_info:
                if not cfg.ddgen_per_table_pid_field:
                    raise ValueError(
                        "Missing ddgen_per_table_pid_field in config for "
                        "database {}".format(dbname))
                ensure_valid_field_name(cfg.ddgen_per_table_pid_field)
                if cfg.ddgen_per_table_pid_field == self.source_hash_fieldname:
                    raise ValueError("Config: ddgen_per_table_pid_field can't "
                                     "be the same as source_hash_fieldname")
            if cfg.ddgen_master_pid_fieldname:
                ensure_valid_field_name(cfg.ddgen_master_pid_fieldname)

        # OK!
        log.debug("Config validated.")

    def encrypt_primary_pid(self, pid: int) -> str:
        """Encrypt a primary PID, producing a RID."""
        if pid is None:  # this is very unlikely!
            raise ValueError("Trying to hash NULL PID!")
            # ... see encrypt_master_pid() below
        return self.primary_pid_hasher.hash(pid)

    def encrypt_master_pid(self, mpid: int) -> Optional[str]:
        """Encrypt a master PID, producing a master RID."""
        if mpid is None:
            return None
            # potentially: risk of revealing the hash
            # and DEFINITELY: two patients, both with no NHS number, must not
            # be equated on the hash (e.g. hash(None) -> hash("None") -> ...)!
        return self.master_pid_hasher.hash(mpid)

    def hash_object(self, l: Any) -> str:
        """
        Hashes a list with Python's built-in hash function.

        We could use Python's build-in hash() function, which produces a 64-bit
        unsigned integer (calculated from: sys.maxint).
        However, there is an outside chance that someone uses a single-field
        table and therefore that this is vulnerable to content discovery via a
        dictionary attack. Thus, we should use a better version.
        """
        return self.change_detection_hasher.hash(repr(l))

    def get_extra_hasher(self, hasher_name: str) -> GenericHasher:
        if hasher_name not in self.extra_hashers.keys():
            raise ValueError(
                "Extra hasher {} requested but doesn't exist; check you have "
                "listed it in 'extra_hash_config_sections' in the config "
                "file".format(hasher_name))
        return self.extra_hashers[hasher_name]

    def get_source_db_names(self) -> List[str]:
        return self.source_db_names

    def set_echo(self, echo: bool) -> None:
        self._echo = echo
        self.admindb.engine.echo = echo
        self.destdb.engine.echo = echo
        for db in self.sources.values():
            db.engine.echo = echo
        # Now, SQLAlchemy will mess things up by adding an additional handler.
        # So, bye-bye:
        for logname in ('sqlalchemy.engine.base.Engine',
                        'sqlalchemy.engine.base.OptionEngine'):
            logger = logging.getLogger(logname)
            # log.critical(logger.__dict__)
            remove_all_logger_handlers(logger)

    def get_src_dialect(self, src_db) -> Any:
        return self.src_dialects[src_db]

    def get_dest_dialect(self) -> Any:
        return self.dest_dialect

    def commit_dest_db(self) -> None:
        self._destdb_transaction_limiter.commit()

    def notify_src_bytes_read(self, n_bytes: int) -> None:
        self._src_bytes_read += n_bytes

    def notify_dest_db_transaction(self, n_rows: int, n_bytes: int) -> None:
        self._destdb_transaction_limiter.notify(n_rows=n_rows, n_bytes=n_bytes)
        # ... may trigger a commit
        self._dest_bytes_written += n_bytes

    def extract_text_extension_permissible(self, extension: str) -> bool:
        if not self.extract_text_extensions_case_sensitive:
            extension = extension.upper()
        if self.extract_text_extensions_permitted:
            return extension in self.extract_text_extensions_permitted
        return extension not in self.extract_text_extensions_prohibited



#!/usr/bin/env python
# crate_anon/tools/merge_csv.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import argparse
import csv
import logging
import sys
from typing import List, TextIO

from crate_anon.common.logsupport import main_only_quicksetup_rootlogger

log = logging.getLogger(__name__)


def merge_csv(filenames: List[str],
              outfile: TextIO = sys.stdout,
              input_dialect: str = 'excel',
              output_dialect: str = 'excel',
              debug: bool = False,
              headers: bool = True) -> None:
    writer = csv.writer(outfile, dialect=output_dialect)
    written_header = False
    header_items = []  # type: List(str)
    for filename in filenames:
        log.info("Processing file " + repr(filename))
        with open(filename, 'r') as f:
            reader = csv.reader(f, dialect=input_dialect)
            if headers:
                if not written_header:
                    header_items = next(reader)
                    if debug:
                        log.debug("Header row: {}".format(repr(header_items)))
                    writer.writerow(header_items)
                    written_header = True
                else:
                    new_headers = next(reader)
                    if new_headers != header_items:
                        raise ValueError(
                            "Header line in file {filename} doesn't match - "
                            "it was {new} but previous was {old}".format(
                                filename=repr(filename),
                                new=repr(new_headers),
                                old=repr(header_items),
                            ))
                    if debug:
                        log.debug("Header row matches previous")
            else:
                if debug:
                    log.debug("No headers in use")
            for row in reader:
                if debug:
                    log.debug("Data row: {}".format(repr(row)))
                writer.writerow(row)


def main():
    main_only_quicksetup_rootlogger()

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "filenames",
        nargs="+",
        help="Names of CSV/TSV files to merge"
    )
    parser.add_argument(
        "--outfile",
        default="-",
        help="Specify an output filename. If omitted or '-', stdout is used.",
    )
    parser.add_argument(
        "--inputdialect",
        default="excel",
        help="The input files' CSV/TSV dialect. Default: %(default)s.",
        choices=csv.list_dialects(),
    )
    parser.add_argument(
        "--outputdialect",
        default="excel",
        help="The output file's CSV/TSV dialect. Default: %(default)s.",
        choices=csv.list_dialects(),
    )
    parser.add_argument(
        "--noheaders",
        action="store_true",
        help="By default, files are assumed to have column headers. "
             "Specify this option to assume no headers.",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Verbose debugging output.",
    )
    progargs = parser.parse_args()

    kwargs = {
        "filenames": progargs.filenames,
        "input_dialect": progargs.inputdialect,
        "output_dialect": progargs.outputdialect,
        "debug": progargs.debug,
        "headers": not progargs.noheaders,
    }
    if progargs.outfile == '-':
        log.info("Writing to stdout")
        merge_csv(outfile=sys.stdout, **kwargs)
    else:
        log.info("Writing to " + repr(progargs.outfile))
        with open(progargs.outfile, 'w') as outfile:
            # noinspection PyTypeChecker
            merge_csv(outfile=outfile, **kwargs)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# crate_anon/crateweb/research/sql_writer.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import logging
from typing import List, Optional

from pyparsing import ParseResults

from crate_anon.common.logsupport import main_only_quicksetup_rootlogger
from crate_anon.common.sql import (
    ColumnId,
    get_first_from_table,
    JoinInfo,
    parser_add_result_column,
    parser_add_from_tables,
    set_distinct_within_parsed,
    TableId,
    WhereCondition,
)
from crate_anon.common.sql_grammar import (
    format_sql,
    SqlGrammar,
    text_from_parsed,
)
from crate_anon.common.sql_grammar_factory import DIALECT_MYSQL, make_grammar
from crate_anon.crateweb.research.research_db_info import (
    research_database_info,
)

log = logging.getLogger(__name__)


# =============================================================================
# Automagically create/manipulate SQL statements based on our extra knowledge
# of the fields that can be used to link across tables/databases.
# =============================================================================

def get_join_info(grammar: SqlGrammar,
                  parsed: ParseResults,
                  jointable: TableId,
                  magic_join: bool = False,
                  nonmagic_join_type: str = "INNER JOIN",
                  nonmagic_join_condition: str = '') -> List[JoinInfo]:
    # Returns e.g. ["INNER JOIN", "tablename", "WHERE somecondition"].
    # INNER JOIN etc. is part of ANSI SQL
    first_from_table = get_first_from_table(parsed)
    from_table_in_join_schema = get_first_from_table(
        parsed,
        match_db=jointable.db(),
        match_schema=jointable.schema())
    exact_match_table = get_first_from_table(
        parsed,
        match_db=jointable.db(),
        match_schema=jointable.schema(),
        match_table=jointable.table())

    if not first_from_table:
        # No tables in query yet.
        # This should not happen; this function is to help with adding
        # new FROM tables to existing FROM clauses.
        log.warning("get_join_info: no tables in query")
        return []

    if exact_match_table:
        # This table is already in the query. No JOIN should be required.
        # log.critical("get_join_info: same table already in query")
        return []

    if not magic_join:
        # log.critical("get_join_info: non-magic join")
        return [JoinInfo(join_type=nonmagic_join_type,
                         table=jointable.identifier(grammar),
                         join_condition=nonmagic_join_condition)]

    if from_table_in_join_schema:
        # Another table from the same database is present. Link on the
        # TRID field.
        # log.critical("get_join_info: joining to another table in same DB")
        return [JoinInfo(
            join_type='INNER JOIN',
            table=jointable.identifier(grammar),
            join_condition="ON {new} = {existing}".format(
                new=research_database_info.get_trid_column(
                    jointable).identifier(grammar),
                existing=research_database_info.get_trid_column(
                    from_table_in_join_schema).identifier(grammar),
            )
        )]

    # OK. So now we're building a cross-database join.
    existing_family = research_database_info.get_db_rid_family(
        first_from_table.schema_id())
    new_family = research_database_info.get_db_rid_family(
        jointable.schema_id())
    # log.critical("existing_family={}, new_family={}".format(
    #     existing_family, new_family))
    if existing_family and existing_family == new_family:
        # log.critical("get_join_info: new DB, same RID family")
        return [JoinInfo(
            join_type='INNER JOIN',
            table=jointable.identifier(grammar),
            join_condition="ON {new} = {existing}".format(
                new=research_database_info.get_rid_column(
                    jointable).identifier(grammar),
                existing=research_database_info.get_rid_column(
                    first_from_table).identifier(grammar),
            )
        )]

    # If we get here, we have to do a complicated join via the MRID.
    # log.critical("get_join_info: new DB, different RID family, using MRID")
    existing_mrid_column = research_database_info.get_mrid_column_from_table(
        first_from_table)
    existing_mrid_table = existing_mrid_column.table_id()
    if not existing_mrid_table:
        raise ValueError(
            "No MRID table available (in the same database as table {}; "
            "cannot link)".format(first_from_table))
    new_mrid_column = research_database_info.get_mrid_column_from_table(
        jointable)
    new_mrid_table = new_mrid_column.table_id()
    existing_mrid_table_in_query = bool(get_first_from_table(
        parsed,
        match_db=existing_mrid_table.db(),
        match_schema=existing_mrid_table.schema(),
        match_table=existing_mrid_table.table()))

    joins = []
    if not existing_mrid_table_in_query:
        joins.append(JoinInfo(
            join_type='INNER JOIN',
            table=existing_mrid_table.identifier(grammar),
            join_condition="ON {m1_trid1} = {t1_trid1}".format(
                m1_trid1=research_database_info.get_trid_column(
                    existing_mrid_table).identifier(grammar),
                t1_trid1=research_database_info.get_trid_column(
                    first_from_table).identifier(grammar),
            )
        ))
    joins.append(JoinInfo(
        join_type='INNER JOIN',
        table=new_mrid_table.identifier(grammar),
        join_condition="ON {m2_mrid2} = {m1_mrid1}".format(
            m2_mrid2=new_mrid_column.identifier(grammar),
            m1_mrid1=existing_mrid_column.identifier(grammar),
        )
    ))
    if jointable != new_mrid_table:
        joins.append(JoinInfo(
            join_type='INNER JOIN',
            table=jointable.identifier(grammar),
            join_condition="ON {t2_trid2} = {m2_trid2}".format(
                t2_trid2=research_database_info.get_trid_column(
                    jointable).identifier(grammar),
                m2_trid2=research_database_info.get_trid_column(
                    new_mrid_table).identifier(grammar),
            )
        ))
    return joins


class SelectElement(object):
    def __init__(self,
                 column_id: ColumnId = None,
                 raw_select: str = '',
                 from_table_for_raw_select: TableId = None,
                 alias: str = ''):
        self.column_id = column_id
        self.raw_select = raw_select
        self.from_table_for_raw_select = from_table_for_raw_select
        self.alias = alias

    def __repr__(self) -> str:
        return (
            "<{qualname}("
            "column_id={column_id}, "
            "raw_select={raw_select}, "
            "from_table_for_raw_select={from_table_for_raw_select}, "
            "alias={alias}) "
            "at {addr}>".format(
                qualname=self.__class__.__qualname__,
                column_id=repr(self.column_id),
                raw_select=repr(self.raw_select),
                from_table_for_raw_select=repr(self.from_table_for_raw_select),
                alias=repr(self.alias),
                addr=hex(id(self)),
            )
        )

    def sql_select_column(self, grammar: SqlGrammar) -> str:
        result = self.raw_select or self.column_id.identifier(grammar)
        if self.alias:
            result += " AS " + self.alias
        return result

    def from_table(self) -> Optional[TableId]:
        if self.raw_select:
            return self.from_table_for_raw_select
        return self.column_id.table_id()

    def from_table_str(self, grammar: SqlGrammar) -> str:
        table_id = self.from_table()
        if not table_id:
            return ''
        return table_id.identifier(grammar)

    def sql_select_from(self, grammar: SqlGrammar) -> str:
        sql = "SELECT " + self.sql_select_column(grammar=grammar)
        from_table = self.from_table()
        if from_table:
            sql += " FROM " + from_table.identifier(grammar)
        return sql


def reparse_select(p: ParseResults, grammar: SqlGrammar) -> ParseResults:
    """
    Internal function for when we get desperate trying to hack around
    the results of pyparsing's efforts.
    """
    return grammar.get_select_statement().parseString(
        text_from_parsed(p, formatted=False),
        parseAll=True
    )


def add_to_select(sql: str,
                  grammar: SqlGrammar,
                  select_elements: List[SelectElement] = None,
                  where_conditions: List[WhereCondition] = None,
                  # For SELECT:
                  distinct: bool = None,  # True, False, or None to leave as is
                  # For WHERE:
                  where_type: str = "AND",
                  bracket_where: bool = False,
                  # For either, for JOIN:
                  magic_join: bool = True,
                  join_type: str = "NATURAL JOIN",
                  join_condition: str = '',
                  # General:
                  formatted: bool = True,
                  debug: bool = False,
                  debug_verbose: bool = False) -> str:
    """
    This function encapsulates our query builder's common operations.
    One premise is that SQL parsing is relatively slow, so we should do this
    only once. We parse; add bits to the parsed structure as required; then
    re-convert to text.

    If you specify table/column, elements will be added to SELECT and FROM
    unless they already exist.

    If you specify where_expression, elements will be added to WHERE.
    In this situation, you should also specify where_table; if the where_table
    isn't yet in the FROM clause, this will be added as well.

    Parsing is SLOW, so we should do as much as possible in a single call to
    this function.
    """
    select_elements = select_elements or []  # type: List[SelectElement]
    where_conditions = where_conditions or []  # type: List[WhereCondition]
    if debug:
        log.info("START: {}".format(sql))
        log.debug("select_elements: {}".format(select_elements))
        log.debug("where_conditions: {}".format(where_conditions))
        log.debug("where_type: {}".format(where_type))
        log.debug("join_type: {}".format(join_type))
        log.debug("join_condition: {}".format(join_condition))

    # -------------------------------------------------------------------------
    # Get going. We have to handle a fresh SQL statement in a slightly
    # different way.
    # -------------------------------------------------------------------------
    if not sql:
        if not select_elements:
            raise ValueError("Fresh SQL statements must include a SELECT "
                             "element")
        # ---------------------------------------------------------------------
        # Fresh SQL statement
        # ---------------------------------------------------------------------
        first_select = select_elements[0]
        select_elements = select_elements[1:]
        sql = first_select.sql_select_from(grammar)
        # log.debug("Starting SQL from scratch as: " + sql)

    # -------------------------------------------------------------------------
    # Parse what we have (which is now, at a minimum, SELECT ... FROM ...).
    # -------------------------------------------------------------------------
    p = grammar.get_select_statement().parseString(sql, parseAll=True)
    if debug and debug_verbose:
        log.debug("start dump:\n" + p.dump())

    existing_tables = p.join_source.from_tables.asList()  # type: List[str]
    new_tables = []  # type: List[TableId]

    def add_new_table(_table_id: TableId) -> None:
        if (_table_id and
                _table_id not in new_tables and
                _table_id.identifier(grammar) not in existing_tables):
            new_tables.append(_table_id)

    # -------------------------------------------------------------------------
    # DISTINCT?
    # -------------------------------------------------------------------------
    if distinct is True:
        set_distinct_within_parsed(p, action='set')
    elif distinct is False:
        set_distinct_within_parsed(p, action='clear')

    # -------------------------------------------------------------------------
    # Process all the (other?) SELECT clauses
    # -------------------------------------------------------------------------
    for se in select_elements:
        p = parser_add_result_column(p, se.sql_select_column(grammar),
                                     grammar=grammar)
        add_new_table(se.from_table())

    # -------------------------------------------------------------------------
    # Process all the WHERE clauses
    # -------------------------------------------------------------------------
    for wc in where_conditions:
        where_expression = wc.sql(grammar)
        if bracket_where:
            where_expression = '(' + where_expression + ')'

        # The tricky bit: inserting it.
        # We use the [0] to overcome the effects of defining these things
        # as a pyparsing Group(), which encapsulates the results in a list.
        if p.where_clause:
            cond = grammar.get_expr().parseString(where_expression,
                                                  parseAll=True)[0]
            extra = [where_type, cond]
            p.where_clause.where_expr.extend(extra)
        else:
            # No WHERE as yet
            # Doing this properly is a nightmare.
            # It's hard to add a *named* ParseResults element to another.
            # So it's very hard to alter p.where_clause.where_expr such that
            # we can continue adding more WHERE clauses if we want.
            # This is the inefficient, cop-out method:
            # (1) Add as plain text
            p.where_clause.append("WHERE " + where_expression)
            # (2) Reparse...
            p = reparse_select(p, grammar=grammar)

        add_new_table(wc.table_id())

    # -------------------------------------------------------------------------
    # Process all the FROM clauses, autojoining as necessary
    # -------------------------------------------------------------------------

    for table_id in new_tables:
        p = parser_add_from_tables(
            p,
            get_join_info(grammar=grammar,
                          parsed=p,
                          jointable=table_id,
                          magic_join=magic_join,
                          nonmagic_join_type=join_type,
                          nonmagic_join_condition=join_condition),
            grammar=grammar)

    if debug and debug_verbose:
        log.debug("end dump:\n" + p.dump())
    result = text_from_parsed(p, formatted=False)
    if formatted:
        result = format_sql(result)
    if debug:
        log.info("END: {}".format(result))
    return result


# =============================================================================
# Unit tests
# =============================================================================

def unit_tests() -> None:
    grammar = make_grammar(DIALECT_MYSQL)
    log.info(add_to_select(
        "SELECT t1.a, t1.b FROM t1 WHERE t1.col1 > 5",
        grammar=grammar,
        select_elements=[SelectElement(
            column_id=ColumnId(table="t2", column="c")
        )],
        magic_join=False  # magic_join requires DB knowledge hence Django
    ))
    log.info(add_to_select(
        "SELECT t1.a, t1.b FROM t1 WHERE t1.col1 > 5",
        grammar=grammar,
        select_elements=[SelectElement(
            column_id=ColumnId(table="t1", column="a")
        )]
    ))
    log.info(add_to_select(
        "",
        grammar=grammar,
        select_elements=[SelectElement(
            column_id=ColumnId(table="t2", column="c")
        )]
    ))
    log.info(add_to_select(
        "SELECT t1.a, t1.b FROM t1 WHERE t1.col1 > 5",
        grammar=grammar,
        where_conditions=[WhereCondition(raw_sql="t1.col2 < 3")]
    ))
    log.info(add_to_select(
        "SELECT t1.a, t1.b FROM t1",
        grammar=grammar,
        where_conditions=[WhereCondition(raw_sql="t1.col1 > 5")]
    ))
    log.info(add_to_select(
        "SELECT t1.a, t1.b FROM t1 WHERE t1.col1 > 5 AND t3.col99 = 100",
        grammar=grammar,
        where_conditions=[WhereCondition(raw_sql="t1.col2 < 3")]
    ))

    # Multiple WHEREs where before there were none:
    log.info(add_to_select(
        "SELECT t1.a, t1.b FROM t1",
        grammar=grammar,
        where_conditions=[WhereCondition(raw_sql="t1.col1 > 99"),
                          WhereCondition(raw_sql="t1.col2 < 999")]
    ))


if __name__ == '__main__':
    main_only_quicksetup_rootlogger()
    unit_tests()



#!/usr/bin/env python
# crate_anon/crateweb/research/models.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

from collections import OrderedDict
import contextlib
import datetime
import io
import logging
from typing import Any, Dict, Generator, List, Iterable, Optional, Tuple, Type
import zipfile

from django.db import connections, DatabaseError, models
from django.db.models import QuerySet
from django.conf import settings
from django.http.request import HttpRequest
from django.db.backends.utils import CursorWrapper
from openpyxl import Workbook

from crate_anon.common.hash import hash64
from crate_anon.common.jsonfunc import (
    JsonClassField,
    json_encode,
    METHOD_STRIP_UNDERSCORE,
    register_for_json,
)
from crate_anon.common.lang import add_info_to_exception, simple_repr
from crate_anon.common.sql import (
    ColumnId,
    columns_to_table_column_hierarchy,
    escape_percent_for_python_dbapi,
    make_grammar,
    sql_string_literal,
    SqlArgsTupleType,
    TableId,
    translate_sql_qmark_to_percent,
    WhereCondition,
)
from crate_anon.common.sql_grammar import format_sql, SqlGrammar
from crate_anon.crateweb.core.dbfunc import (
    dictfetchall,
    get_fieldnames_from_cursor,
    make_tsv_row,
)
from crate_anon.crateweb.extra.excel import excel_to_bytes
from crate_anon.crateweb.research.html_functions import (
    highlight_text,
    HtmlElementCounter,
    N_CSS_HIGHLIGHT_CLASSES,
    prettify_sql_html,
)
from crate_anon.crateweb.research.research_db_info import (
    research_database_info,
)
from crate_anon.crateweb.research.sql_writer import (
    add_to_select,
    SelectElement,
)

log = logging.getLogger(__name__)


# =============================================================================
# Hacking django-pyodbc-azure, to stop it calling cursor.nextset() every time
# you ask it to do cursor.fetchone()
# =============================================================================

DJANGO_PYODBC_AZURE_ENGINE = 'sql_server.pyodbc'


def replacement_sqlserver_pyodbc_cursorwrapper_fetchone(self) -> List[Any]:
    # To replace CursorWrapper.fetchone() in sql_server/pyodbc/base.py
    #
    # log.critical("Using monkeypatched fetchone(); self: {}; self.cursor: "
    #              "{}".format(repr(self), repr(self.cursor)))
    row = self.cursor.fetchone()
    if row is not None:
        row = self.format_row(row)
    # BUT DO NOT CALL self.cursor.nextset()
    return row


def hack_django_pyodbc_azure_cursorwrapper() -> None:
    # I thought I wanted to modify an INSTANCE, not a CLASS.
    # - https://tryolabs.com/blog/2013/07/05/run-time-method-patching-python/
    # To modify a class, we do
    #       SomeClass.method = newmethod
    # But to modify an instance, we use
    #       instance.method = types.MethodType(newmethod, instance)
    # However, it turned out the instance was actually part of a long chain
    # of cursor wrappers, including the Django debug toolbar; classes included
    #       debug_toolbar.panels.sql.tracking.NormalCursorWrapper
    #       django.db.backends.utils.CursorDebugWrapper
    # and actually, modifying the class is a sensible thing.
    try:
        # noinspection PyUnresolvedReferences
        from sql_server.pyodbc.base import CursorWrapper
        log.info("Monkey-patching sql_server.pyodbc.base.CursorWrapper."
                 "fetchone to disable automatic call to cursor.nextset()")
        CursorWrapper.fetchone = replacement_sqlserver_pyodbc_cursorwrapper_fetchone  # noqa
    except ImportError:
        return


if getattr(settings, 'DISABLE_DJANGO_PYODBC_AZURE_CURSOR_FETCHONE_NEXTSET',
           True):
    # http://stackoverflow.com/questions/5601590/how-to-define-a-default-value-for-a-custom-django-setting  # noqa
    hack_django_pyodbc_azure_cursorwrapper()


# =============================================================================
# Cursors
# =============================================================================

def debug_query() -> None:
    cursor = connections['research'].cursor()
    cursor.execute("SELECT 'debug'")


def get_executed_researchdb_cursor(sql, args: List[Any] = None) -> CursorWrapper:  # noqa
    args = args or []
    cursor = connections['research'].cursor()
    try:
        cursor.execute(sql, args or None)
    except DatabaseError as exception:
        add_info_to_exception(exception, {'sql': sql, 'args': args})
        raise
    # noinspection PyTypeChecker
    return contextlib.closing(cursor)


# =============================================================================
# Query highlighting class
# =============================================================================

HIGHLIGHT_FWD_REF = "Highlight"


class Highlight(models.Model):
    """
    Represents the highlighting of a query.
    """
    id = models.AutoField(primary_key=True)  # automatic
    user = models.ForeignKey(settings.AUTH_USER_MODEL)
    colour = models.PositiveSmallIntegerField(verbose_name="Colour number")
    text = models.CharField(max_length=255, verbose_name="Text to highlight")
    active = models.BooleanField(default=True)

    def __str__(self) -> str:
        return "colour={}, text={}".format(self.colour, self.text)

    def get_safe_colour(self) -> int:
        if self.colour is None:
            return 0
        return min(self.colour, N_CSS_HIGHLIGHT_CLASSES - 1)

    @staticmethod
    def as_ordered_dict(highlight_list) -> Dict[int, List[HIGHLIGHT_FWD_REF]]:
        d = dict()
        for highlight in highlight_list:
            n = highlight.get_safe_colour()
            if n not in d:
                d[n] = []  # type: List[HIGHLIGHT_FWD_REF]
            d[n].append(highlight)
        return OrderedDict(sorted(d.items()))

    @staticmethod
    def get_active_highlights(request: HttpRequest) -> QuerySet:
        return Highlight.objects.filter(user=request.user, active=True)

    def activate(self) -> None:
        self.active = True
        self.save()

    def deactivate(self) -> None:
        self.active = False
        self.save()


# =============================================================================
# Query class
# =============================================================================

QUERY_FWD_REF = "Query"


class Query(models.Model):
    """
    Class to query the research database.
    """
    class Meta:
        app_label = "research"

    id = models.AutoField(primary_key=True)  # automatic
    user = models.ForeignKey(settings.AUTH_USER_MODEL)

    sql = models.TextField(verbose_name='SQL query')
    sql_hash = models.BigIntegerField(
        verbose_name='64-bit non-cryptographic hash of SQL query')
    args = JsonClassField(verbose_name='SQL arguments (as JSON)', null=True)
    # ... https://github.com/shrubberysoft/django-picklefield
    raw = models.BooleanField(
        default=False, verbose_name='SQL is raw, not parameter-substituted')
    qmark = models.BooleanField(
        default=True,
        verbose_name='Parameter-substituted SQL uses ?, not %s, '
        'as placeholders')
    active = models.BooleanField(default=True)  # see save() below
    created = models.DateTimeField(auto_now_add=True)
    deleted = models.BooleanField(
        default=False,
        verbose_name="Deleted from the user's perspective. "
                     "Audited queries are never properly deleted.")
    audited = models.BooleanField(default=False)

    def __repr__(self) -> str:
        return simple_repr(self, ['id', 'user', 'sql', 'args', 'raw', 'qmark',
                                  'active', 'created', 'deleted', 'audited'])

    def save(self, *args, **kwargs) -> None:
        """
        Custom save method.
        Ensures that only one Query has active == True for a given user.
        Also sets the hash.
        """
        # http://stackoverflow.com/questions/1455126/unique-booleanfield-value-in-django  # noqa
        if self.active:
            Query.objects.filter(user=self.user, active=True)\
                         .update(active=False)
        self.sql_hash = hash64(self.sql)
        super().save(*args, **kwargs)

    # -------------------------------------------------------------------------
    # Fetching
    # -------------------------------------------------------------------------

    @staticmethod
    def get_active_query_or_none(request: HttpRequest) \
            -> Optional[QUERY_FWD_REF]:
        if not request.user.is_authenticated():
            return None
        try:
            return Query.objects.get(user=request.user, active=True)
        except Query.DoesNotExist:
            return None

    @staticmethod
    def get_active_query_id_or_none(request: HttpRequest) -> Optional[int]:
        if not request.user.is_authenticated():
            return None
        try:
            query = Query.objects.get(user=request.user, active=True)
            return query.id
        except Query.DoesNotExist:
            return None

    # -------------------------------------------------------------------------
    # Activating, deleting, auditing
    # -------------------------------------------------------------------------

    def activate(self) -> None:
        self.active = True
        self.save()

    def mark_audited(self) -> None:
        if self.audited:
            return
        self.audited = True
        self.save()

    def mark_deleted(self) -> None:
        if self.deleted:
            # log.debug("pointless")
            return
        self.deleted = True
        self.active = False
        # log.debug("about to save")
        self.save()
        # log.debug("saved")

    def delete_if_permitted(self) -> None:
        """If a query has been audited, it isn't properly deleted."""
        if self.deleted:
            log.debug("already flagged as deleted")
            return
        if self.audited:
            log.debug("marking as deleted")
            self.mark_deleted()
        else:
            # actually delete
            log.debug("actually deleting")
            self.delete()

    def audit(self, count_only: bool = False, n_records: int = 0,
              failed: bool = False, fail_msg: str = "") -> None:
        a = QueryAudit(query=self,
                       count_only=count_only,
                       n_records=n_records,
                       failed=failed,
                       fail_msg=fail_msg)
        a.save()
        self.mark_audited()

    # -------------------------------------------------------------------------
    # Highlights
    # -------------------------------------------------------------------------

    def add_highlight(self, text: str, colour: int = 0) -> None:
        h = Highlight(text=text, colour=colour)
        self.highlight_set.add(h)

    def get_highlights_as_dict(self) -> Dict[int, Iterable[Highlight]]:
        d = OrderedDict()
        for n in range(N_CSS_HIGHLIGHT_CLASSES):
            d[n] = Highlight.objects.filter(query_id=self.id, colour=n)
        return d

    def get_highlight_descriptions(self) -> List[str]:
        d = self.get_highlights_as_dict()
        desc = []
        for n in range(N_CSS_HIGHLIGHT_CLASSES):
            if d[n]:
                # noinspection PyTypeChecker
                desc.append(", ".join(highlight_text(h.text, n) for h in d[n]))
        return desc

    # -------------------------------------------------------------------------
    # SQL queries
    # -------------------------------------------------------------------------

    def get_original_sql(self) -> str:
        # noinspection PyTypeChecker
        return self.sql

    def get_sql_args_for_django(self) -> Tuple[str, Optional[List[Any]]]:
        """
        Get sql/args in a format suitable for Django, with %s placeholders,
        or as escaped raw SQL.
        """
        if self.raw:
            # noinspection PyTypeChecker
            sql = escape_percent_for_python_dbapi(self.sql)
            args = None
        else:
            if self.qmark:
                # noinspection PyTypeChecker
                sql = translate_sql_qmark_to_percent(self.sql)
            else:
                sql = self.sql
            args = self.args
        return sql, args

    def get_executed_cursor(self, sql_append_raw: str = None) -> CursorWrapper:
        """
        Get cursor with a query executed
        """
        (sql, args) = self.get_sql_args_for_django()
        if sql_append_raw:
            sql += sql_append_raw
        return get_executed_researchdb_cursor(sql, args)

    # def gen_rows(self,
    #              firstrow: int = 0,
    #              lastrow: int = None) -> Generator[List[Any], None, None]:
    #     """
    #     Generate rows from the query.
    #     """
    #     if firstrow > 0 or lastrow is not None:
    #         sql_append_raw = " LIMIT {f},{n}".format(
    #             f=firstrow,
    #             n=(lastrow - firstrow + 1),
    #         )
    #         # zero-indexed;
    #         # http://dev.mysql.com/doc/refman/5.0/en/select.html
    #     else:
    #         sql_append_raw = None
    #     with self.get_executed_cursor(sql_append_raw) as cursor:
    #         row = cursor.fetchone()
    #         while row is not None:
    #             yield row
    #             row = cursor.fetchone()

    def make_tsv(self) -> str:
        with self.get_executed_cursor() as cursor:
            fieldnames = get_fieldnames_from_cursor(cursor)
            tsv = make_tsv_row(fieldnames)
            row = cursor.fetchone()
            while row is not None:
                tsv += make_tsv_row(row)
                row = cursor.fetchone()
        return tsv

    def make_excel(self) -> bytes:
        wb = Workbook()
        wb.remove_sheet(wb.active)  # remove the autocreated blank sheet
        sheetname = "query_{}".format(self.id)
        ws = wb.create_sheet(sheetname)
        now = datetime.datetime.now()
        with self.get_executed_cursor() as cursor:
            fieldnames = get_fieldnames_from_cursor(cursor)
            ws.append(fieldnames)
            row = cursor.fetchone()
            while row is not None:
                ws.append(tuple(row))
                # - openpyxl doesn't believe in duck-typing; see
                #   openpyxl/worksheet/worksheet.py
                # - Sometimes this works (e.g. from MySQL), but sometimes it
                #   fails, e.g. when the row is of type pyodbc.Row
                # - So we must coerce to list or tuple
                row = cursor.fetchone()
                # BUG in django-pyodbc-azure==1.10.4.0 (providing
                # sql_server/*), 2017-02-17: this causes
                # ProgrammingError "No results. Previous SQL was not a query."
                # The problem relates to sql_server/pyodbc/base.py
                # CursorWrapper.fetchone() calling self.cursor.nextset(); if
                # you comment this out, it works fine.
                # Related:
                # - https://github.com/pymssql/pymssql/issues/98

        sql_ws = wb.create_sheet(title="SQL")
        sql_ws.append(["SQL", "Executed_at"])
        sql_ws.append([self.get_original_sql(), now])
        return excel_to_bytes(wb)

    def dictfetchall(self) -> List[Dict[str, Any]]:
        """Generates all results as a list of OrderedDicts."""
        with self.get_executed_cursor() as cursor:
            return dictfetchall(cursor)


# =============================================================================
# Query auditing class
# =============================================================================

class QueryAudit(models.Model):
    """
    Audit log for a query.
    """
    id = models.AutoField(primary_key=True)  # automatic
    query = models.ForeignKey('Query')
    when = models.DateTimeField(auto_now_add=True)
    count_only = models.BooleanField(default=False)
    n_records = models.IntegerField(default=0)
    # ... not PositiveIntegerField; SQL Server gives -1, for example
    failed = models.BooleanField(default=False)
    fail_msg = models.TextField()

    def __str__(self):
        return "<QueryAudit id={}>".format(self.id)


# =============================================================================
# Lookup class for secret RID-to-PID conversion
# =============================================================================

class PidLookupRouter(object):
    # https://docs.djangoproject.com/en/1.8/topics/db/multi-db/
    # https://newcircle.com/s/post/1242/django_multiple_database_support
    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def db_for_read(self, model: Type[models.Model], **hints) -> Optional[str]:
        """
        read model PidLookup -> look at database secret
        """
        # log.debug("PidLookupRouter: {}".format(model._meta.model_name))
        # if model._meta.model_name == PidLookup._meta.model_name:
        if model == PidLookup:
            return 'secret'
        return None

    # noinspection PyUnusedLocal
    @staticmethod
    def allow_migrate(db, app_label, model_name=None, **hints):
        # 2017-02-12, to address bug:
        # - https://code.djangoproject.com/ticket/27054
        # See also:
        # - https://docs.djangoproject.com/en/1.10/topics/db/multi-db/#using-other-management-commands  # noqa
        return db == 'default'


class PidLookup(models.Model):
    """
    Lookup class for secret RID-to-PID conversion.
    Uses the 'secret' database connection.

    Use as e.g. Lookup(pid=XXX)
    """
    pid = models.PositiveIntegerField(
        primary_key=True,
        db_column=settings.SECRET_MAP['PID_FIELD'])
    mpid = models.PositiveIntegerField(
        db_column=settings.SECRET_MAP['MASTER_PID_FIELD'])
    rid = models.CharField(
        db_column=settings.SECRET_MAP['RID_FIELD'],
        max_length=settings.SECRET_MAP['MAX_RID_LENGTH'])
    mrid = models.CharField(
        db_column=settings.SECRET_MAP['MASTER_RID_FIELD'],
        max_length=settings.SECRET_MAP['MAX_RID_LENGTH'])
    trid = models.PositiveIntegerField(
        db_column=settings.SECRET_MAP['TRID_FIELD'])

    class Meta:
        managed = False
        db_table = settings.SECRET_MAP['TABLENAME']


def get_pid_lookup(trid: int = None,
                   rid: str = None,
                   mrid: str = None) -> PidLookup:
    if trid is not None:
        lookup = PidLookup.objects.get(trid=trid)
    elif rid is not None:
        lookup = PidLookup.objects.get(rid=rid)
    elif mrid is not None:
        lookup = PidLookup.objects.get(mrid=mrid)
    else:
        raise ValueError("no input")
    return lookup


def get_mpid(trid: int = None,
             rid: str = None,
             mrid: str = None) -> int:
    lookup = get_pid_lookup(trid=trid, rid=rid, mrid=mrid)
    # noinspection PyTypeChecker
    return lookup.mpid


def get_pid(trid: int = None,
            rid: str = None,
            mrid: str = None) -> int:
    lookup = get_pid_lookup(trid=trid, rid=rid, mrid=mrid)
    # noinspection PyTypeChecker
    return lookup.pid


# =============================================================================
# Patient Explorer multi-query class
# =============================================================================

"""

1. Patient ID query

- Single database is easy; we can use RID or TRID, and therefore TRID for
  performance.
  Note that UNION gives only DISTINCT results by default ("UNION ALL" gives
  everything).
  ... http://stackoverflow.com/questions/49925/what-is-the-difference-between-union-and-union-all

    -- Clear, but extensibility of boolean logic less clear:
    SELECT trid
        FROM diagnosis_table
        WHERE diagnosis LIKE 'F20%'
    INTERSECT
    SELECT trid
        FROM progress_note_table
        WHERE note LIKE '%schizophreni%' OR note LIKE '%depression%'
    ORDER BY trid
    ... logic across tables requires careful arrangement of UNION vs. INTERSECT
    ... logic for multiple fields within one table can be done with AND/OR

    -- Slower (?), but simpler to manipulate logic?
    SELECT DISTINCT something.trid
    FROM diagnosis_table INNER JOIN progress_note_table
    ON diagnosis_table.trid = progress_note_table.trid
    WHERE
        diagnosis_table.diagnosis LIKE 'F20%'
        AND (progress_note_table.note LIKE '%schizophreni%'
             OR progress_note_table.notenote LIKE '%depression%')
    ORDER BY something.trid
    -- ... boolean logic can all be encapsulated in a single WHERE clause
    -- ... can also share existing join code
    -- ... ?reasonable speed since the TRID fields will be indexed
    -- ... preferable.

1b. Which ID for the patient ID query

    ... the TRID (for speed, inc. sorting) of the first database
    ... can use the TRID from the first "where clause" table
        (don't have to join to a master patient table)
    ... join everything across databases as before

2. Results queries

    -- Something like:

    SELECT rid, date_of_note, note
    FROM progress_note_table
    WHERE trid IN ( ... patient_id_query ... )
    ORDER BY trid

    SELECT rid, date_of_diagnosis, diagnosis, diagnosis_description
    FROM diagnosis_table
    WHERE trid IN ( ... patient_id_query ... )
    ORDER BY trid


This means we will repeat the patient_id_query, which may be inefficient.
Options:
- store the TRIDs in Python, then pass them as arguments
  ... at which point the SQL string/packet length becomes relevant;
  ... http://stackoverflow.com/questions/1869753/maximum-size-for-a-sql-server-query-in-clause-is-there-a-better-approach
  ... http://stackoverflow.com/questions/16335011/what-is-maximum-query-size-for-mysql
  ... http://stackoverflow.com/questions/96553/practical-limit-to-length-of-sql-query-specifically-mysql
- let the database worry about it
  ... probably best for now!


3. Display

    One patient per page, with multiple results tables.

===========

- Boolean logic on patient selection
    ... within


"""  # noqa


# =============================================================================
# PatientMultiQuery
# =============================================================================

@register_for_json(method=METHOD_STRIP_UNDERSCORE)
class PatientMultiQuery(object):
    def __init__(self,
                 output_columns: List[ColumnId] = None,
                 patient_conditions: List[WhereCondition] = None,
                 manual_patient_id_query: str = ''):
        self._output_columns = output_columns or []  # type: List[ColumnId]
        self._patient_conditions = patient_conditions or []  # type: List[WhereCondition]  # noqa
        self._manual_patient_id_query = manual_patient_id_query or ''

    def __repr__(self) -> str:
        return (
            "<{qualname}("
            "output_columns={output_columns}, "
            "patient_conditions={patient_conditions}, "
            "manual_patient_id_query={manual_patient_id_query}"
            ") at {addr}>".format(
                qualname=self.__class__.__qualname__,
                output_columns=repr(self._output_columns),
                patient_conditions=repr(self._patient_conditions),
                manual_patient_id_query=repr(self._manual_patient_id_query),
                addr=hex(id(self)),
            )
        )

    def __eq__(self, other: 'PatientMultiQuery') -> bool:
        return (
            self._output_columns == other._output_columns and
            self._patient_conditions == other._patient_conditions and
            self._manual_patient_id_query == other._manual_patient_id_query
        )

    def __hash__(self) -> int:
        """
        WARNING: Python's hash() function converts the result of __hash__()
        to the integer width of the host machine, so 64-bit results can get
        down-converted to 32 bits. Use hash64() directly if you want a 64-bit
        result.
        """
        return self.hash64()

    def hash64(self) -> int:
        return hash64(json_encode(self))

    def get_output_columns(self) -> List[ColumnId]:
        return self._output_columns

    def has_output_columns(self) -> bool:
        return bool(self._output_columns)

    def ok_to_run(self) -> bool:
        return self.has_output_columns() and self.has_patient_id_query()

    def get_patient_conditions(self) -> List[WhereCondition]:
        return self._patient_conditions

    def get_manual_patient_id_query(self) -> str:
        return self._manual_patient_id_query

    def add_output_column(self, column_id: ColumnId) -> None:
        if column_id not in self._output_columns:
            self._output_columns.append(column_id)
            self._output_columns.sort()

    def clear_output_columns(self) -> None:
        self._output_columns = []

    def add_patient_condition(self, where: WhereCondition) -> None:
        if where not in self._patient_conditions:
            self._patient_conditions.append(where)
            self._patient_conditions.sort()

    def clear_patient_conditions(self) -> None:
        self._patient_conditions = []

    def set_override_query(self, query: str) -> None:
        self._manual_patient_id_query = query

    def _get_select_mrid_column(self) -> Optional[ColumnId]:
        if not self._patient_conditions:
            return None
        return research_database_info.get_linked_mrid_column(
            self._patient_conditions[0].table_id())

    def has_patient_id_query(self) -> bool:
        if self._manual_patient_id_query:
            return True
        if self._patient_conditions:
            mrid_col = self._get_select_mrid_column()
            if mrid_col and mrid_col.is_valid():
                return True
        return False

    def patient_id_query(self, with_order_by: bool = True) -> str:
        # Returns an SQL SELECT statement based on the list of WHERE conditions
        # already stored, joined with AND by default.

        if self._manual_patient_id_query:
            # User has specified one manually.
            return self._manual_patient_id_query

        if not self._patient_conditions:
            return ''

        grammar = research_database_info.grammar
        select_mrid_column = self._get_select_mrid_column()
        if not select_mrid_column.is_valid():
            log.warning(
                "PatientMultiQuery.patient_id_query(): invalid "
                "select_mrid_column: {}".format(repr(select_mrid_column)))
            # One way this can happen: (1) a user saves a PMQ; (2) the
            # administrator removes one of the databases!
            return ''
        mrid_alias = "_mrid"
        sql = add_to_select(
            '',
            grammar=grammar,
            select_elements=[SelectElement(column_id=select_mrid_column,
                                           alias=mrid_alias)],
            distinct=True,
            where_conditions=(self._patient_conditions + [WhereCondition(
                column_id=select_mrid_column, op="IS NOT NULL")]),
            where_type="AND",
            magic_join=True,
            formatted=False
        )
        if with_order_by:
            sql += " ORDER BY " + mrid_alias
            # ... ORDER BY is important for consistency across runs
        sql = format_sql(sql)
        # log.critical(sql)
        return sql

    def all_full_queries(self) -> List[Tuple[TableId, str, List[Any]]]:
        return self.all_queries(mrids=None)

    def all_queries_specific_patients(
            self,
            mrids: List[int]) -> List[Tuple[TableId, str, List[Any]]]:
        return self.all_queries(mrids=mrids)

    def all_queries(self,
                    mrids: List[Any] = None) -> List[Tuple[TableId, str,
                                                           List[Any]]]:
        queries = []
        table_columns_map = columns_to_table_column_hierarchy(
            self._output_columns, sort=True)
        for table, columns in table_columns_map:
            table_sql_args = self.make_query(table_id=table,
                                             columns=columns,
                                             mrids=mrids)
            queries.append(table_sql_args)
        return queries

    def where_patient_clause(self, table_id: TableId,
                             grammar: SqlGrammar,
                             mrids: List[Any] = None) -> SqlArgsTupleType:
        """Returns (sql, args)."""
        mrid_column = research_database_info.get_mrid_column_from_table(
            table_id)
        if mrids:
            in_clause = ",".join(["?"] * len(mrids))
            # ... see notes for translate_sql_qmark_to_percent()
            args = mrids
        else:
            # If we haven't specified specific patients, use our patient-
            # finding query.
            in_clause = self.patient_id_query(with_order_by=False)
            # ... SQL Server moans if you use use ORDER BY in a subquery:
            # "The ORDER BY clause is invalid in views, inline functions,
            # derived tables, subqueries, ... unless TOP, OFFSET or FOR XML
            # is specified."
            args = []
        sql = "{mrid} IN ({in_clause})".format(
            mrid=mrid_column.identifier(grammar),
            in_clause=in_clause)
        return sql, args

    def make_query(self,
                   table_id: TableId,
                   columns: List[ColumnId],
                   mrids: List[Any] = None) -> Tuple[TableId, str, List[Any]]:
        if not columns:
            raise ValueError("No columns specified")
        grammar = research_database_info.grammar
        mrid_column = research_database_info.get_mrid_column_from_table(
            table_id)
        all_columns = [mrid_column]
        for c in columns:
            if c not in all_columns:
                all_columns.append(c)
        where_clause, args = self.where_patient_clause(table_id, grammar,
                                                       mrids)
        select_elements = [SelectElement(column_id=col)
                           for col in all_columns]
        where_conditions = [WhereCondition(raw_sql=where_clause)]
        sql = add_to_select('',
                            grammar=grammar,
                            select_elements=select_elements,
                            where_conditions=where_conditions,
                            magic_join=True,
                            formatted=True)
        return table_id, sql, args

    # -------------------------------------------------------------------------
    # Display
    # -------------------------------------------------------------------------

    def output_cols_html(self) -> str:
        grammar = research_database_info.grammar
        return prettify_sql_html("\n".join(
            [column_id.identifier(grammar)
             for column_id in self.get_output_columns()]))

    def pt_conditions_html(self) -> str:
        grammar = research_database_info.grammar
        return prettify_sql_html("\nAND ".join([
            wc.sql(grammar) for wc in self.get_patient_conditions()]))

    def summary_html(self, element_counter: HtmlElementCounter) -> str:
        def collapser(x: str) -> str:
            return element_counter.overflow_div(contents=x)
        outcols = self.output_cols_html()
        manual_query = self.get_manual_patient_id_query()
        if manual_query:
            manual_or_auto = " (MANUAL)"
            ptselect = prettify_sql_html(manual_query)
        else:
            manual_or_auto = ""
            ptselect = self.pt_conditions_html()
        return """
            Output columns:<br>
            {outcols}
            Patient selection:<br>
            {ptselect}
        """.format(
            outcols=collapser(outcols),
            manual_or_auto=manual_or_auto,
            ptselect=collapser(ptselect),
        )

    # -------------------------------------------------------------------------
    # Data finder: COUNT(*) for all patient tables
    # -------------------------------------------------------------------------

    def gen_data_finder_queries(self, mrids: List[Any] = None) \
            -> Generator[Tuple[str, str, List[Any]], None, None]:
        """
        Generates (table_identifier, sql, args).
        When executed, query gives:
            research_id, table_name, n_records, min_date, max_date
        """
        grammar = research_database_info.grammar
        mrid_alias = 'master_research_id'
        table_name_alias = 'table_name'
        n_records_alias = 'n_records'
        min_date_alias = 'min_date'
        max_date_alias = 'max_date'
        for table_id in research_database_info.get_mrid_linkable_patient_tables():  # noqa
            mrid_col = research_database_info.get_mrid_column_from_table(
                table=table_id)
            date_col = research_database_info.get_default_date_column(
                table=table_id)
            if research_database_info.table_contains(table_id, date_col):
                min_date = "MIN({})".format(date_col.identifier(grammar))
                max_date = "MAX({})".format(date_col.identifier(grammar))
            else:
                min_date = "NULL"
                max_date = "NULL"
                # ... OK (at least in MySQL) to do:
                # SELECT col1, COUNT(*), NULL FROM table GROUP BY col1;
            where_clause, args = self.where_patient_clause(
                table_id, grammar, mrids)
            table_identifier = table_id.identifier(grammar)
            select_elements = [
                SelectElement(column_id=mrid_col, alias=mrid_alias),
                SelectElement(raw_select=sql_string_literal(table_identifier),
                              alias=table_name_alias),
                SelectElement(raw_select='COUNT(*)',
                              from_table_for_raw_select=table_id,
                              alias=n_records_alias),
                SelectElement(raw_select=min_date,
                              from_table_for_raw_select=table_id,
                              alias=min_date_alias),
                SelectElement(raw_select=max_date,
                              from_table_for_raw_select=table_id,
                              alias=max_date_alias),
            ]
            where_conditions = [WhereCondition(raw_sql=where_clause)]
            sql = add_to_select('',
                                grammar=grammar,
                                select_elements=select_elements,
                                where_conditions=where_conditions,
                                magic_join=True,
                                formatted=False)
            sql += "\nGROUP BY " + mrid_col.identifier(grammar)
            sql += "\nORDER BY " + mrid_alias
            sql = format_sql(sql)
            yield table_identifier, sql, args

    # -------------------------------------------------------------------------
    # Monster data: SELECT * for all patient tables
    # -------------------------------------------------------------------------

    def gen_monster_queries(self, mrids: List[int] = None) \
            -> Generator[List[Tuple[TableId, str, List[Any]]], None, None]:
        grammar = research_database_info.grammar
        for table_id in research_database_info.get_mrid_linkable_patient_tables():  # noqa
            mrid_col = research_database_info.get_mrid_column_from_table(
                table=table_id)
            where_clause, args = self.where_patient_clause(
                table_id, grammar, mrids)
            # We add the WHERE using our magic query machine, to get the joins
            # right:
            select_elements = [
                SelectElement(raw_select='*',
                              from_table_for_raw_select=table_id),
            ]
            where_conditions = [
                WhereCondition(raw_sql=where_clause,
                               from_table_for_raw_sql=mrid_col.table_id()),
            ]
            sql = add_to_select(
                '',
                grammar=grammar,
                select_elements=select_elements,
                where_conditions=where_conditions,
                magic_join=True,
                formatted=False)
            sql += " ORDER BY " + mrid_col.identifier(grammar)
            sql = format_sql(sql)
            yield table_id, sql, args


# =============================================================================
# PatientExplorer
# =============================================================================

PATIENT_EXPLORER_FWD_REF = "PatientExplorer"


class PatientExplorer(models.Model):
    """
    Class to explore the research database on a per-patient basis.
    """
    class Meta:
        app_label = "research"

    id = models.AutoField(primary_key=True)  # automatic
    user = models.ForeignKey(settings.AUTH_USER_MODEL)
    patient_multiquery = JsonClassField(
        verbose_name='PatientMultiQuery as JSON',
        null=True)  # type: PatientMultiQuery
    pmq_hash = models.BigIntegerField(
        verbose_name='64-bit non-cryptographic hash of JSON of '
                     'patient_multiquery')
    active = models.BooleanField(default=True)  # see save() below
    created = models.DateTimeField(auto_now_add=True)
    deleted = models.BooleanField(
        default=False,
        verbose_name="Deleted from the user's perspective. "
                     "Audited queries are never properly deleted.")
    audited = models.BooleanField(default=False)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if not self.patient_multiquery:
            self.patient_multiquery = PatientMultiQuery()

    def __str__(self) -> str:
        return "<PatientExplorer id={}>".format(self.id)

    def save(self, *args, **kwargs) -> None:
        """
        Custom save method. Ensures that only one PatientExplorer has
        active == True for a given user.
        Also sets the hash.
        """
        if self.active:
            PatientExplorer.objects\
                .filter(user=self.user, active=True)\
                .update(active=False)
        self.pmq_hash = self.patient_multiquery.hash64()
        # Beware: Python's hash() function will downconvert to 32 bits on 32-bit
        # machines; use pmq.hash64() directly, not hash(pmq).
        super().save(*args, **kwargs)

    # -------------------------------------------------------------------------
    # Fetching
    # -------------------------------------------------------------------------

    @staticmethod
    def get_active_pe_or_none(request: HttpRequest) \
            -> Optional[PATIENT_EXPLORER_FWD_REF]:
        if not request.user.is_authenticated():
            return None
        try:
            return PatientExplorer.objects.get(user=request.user, active=True)
        except PatientExplorer.DoesNotExist:
            return None

    @staticmethod
    def get_active_pe_id_or_none(request: HttpRequest) -> Optional[int]:
        if not request.user.is_authenticated():
            return None
        try:
            pe = PatientExplorer.objects.get(user=request.user, active=True)
            return pe.id
        except PatientExplorer.DoesNotExist:
            return None

    # -------------------------------------------------------------------------
    # Activating, deleting, auditing
    # -------------------------------------------------------------------------

    def activate(self) -> None:
        self.active = True
        self.save()

    def mark_audited(self) -> None:
        if self.audited:
            return
        self.audited = True
        self.save()

    def mark_deleted(self) -> None:
        if self.deleted:
            # log.debug("pointless")
            return
        self.deleted = True
        self.active = False
        # log.debug("about to save")
        self.save()
        # log.debug("saved")

    def delete_if_permitted(self) -> None:
        """If a PE has been audited, it isn't properly deleted."""
        if self.deleted:
            log.debug("already flagged as deleted")
            return
        if self.audited:
            log.debug("marking as deleted")
            self.mark_deleted()
        else:
            # actually delete
            log.debug("actually deleting")
            self.delete()

    def audit(self, count_only: bool = False, n_records: int = 0,
              failed: bool = False, fail_msg: str = "") -> None:
        a = PatientExplorerAudit(patient_explorer=self,
                                 count_only=count_only,
                                 n_records=n_records,
                                 failed=failed,
                                 fail_msg=fail_msg)
        a.save()
        self.mark_audited()

    # -------------------------------------------------------------------------
    # Highlights
    # -------------------------------------------------------------------------

    def add_highlight(self, text: str, colour: int = 0) -> None:
        h = Highlight(text=text, colour=colour)
        self.highlight_set.add(h)

    def get_highlights_as_dict(self) -> Dict[int, Iterable[Highlight]]:
        d = OrderedDict()
        for n in range(N_CSS_HIGHLIGHT_CLASSES):
            d[n] = Highlight.objects.filter(query_id=self.id, colour=n)
        return d

    def get_highlight_descriptions(self) -> List[str]:
        d = self.get_highlights_as_dict()
        desc = []
        for n in range(N_CSS_HIGHLIGHT_CLASSES):
            if d[n]:
                # noinspection PyTypeChecker
                desc.append(", ".join(highlight_text(h.text, n) for h in d[n]))
        return desc

    # -------------------------------------------------------------------------
    # Using the internal PatientMultiQuery
    # -------------------------------------------------------------------------

    def all_queries(self,
                    mrids: List[Any] = None) -> List[Tuple[TableId, str,
                                                           List[Any]]]:
        return self.patient_multiquery.all_queries(mrids=mrids)

    @staticmethod
    def get_executed_cursor(sql: str, args: List[Any] = None) -> CursorWrapper:
        """
        Get cursor with a query executed
        """
        sql = translate_sql_qmark_to_percent(sql)
        return get_executed_researchdb_cursor(sql, args)

    def get_patient_mrids(self) -> List[int]:
        sql = self.patient_multiquery.patient_id_query(with_order_by=True)
        # log.critical(sql)
        with self.get_executed_cursor(sql) as cursor:
            return [row[0] for row in cursor.fetchall()]

    def get_zipped_tsv_binary(self) -> bytes:
        # Don't pass giant result sets around beyond what's necessary.
        # Use cursor.fetchone()
        grammar = make_grammar(settings.RESEARCH_DB_DIALECT)
        memfile = io.BytesIO()
        z = zipfile.ZipFile(memfile, "w")
        for table_id, sql, args in self.patient_multiquery.all_queries():
            with self.get_executed_cursor(sql, args) as cursor:
                fieldnames = get_fieldnames_from_cursor(cursor)
                tsv = make_tsv_row(fieldnames)
                row = cursor.fetchone()
                while row is not None:
                    tsv += make_tsv_row(row)
                    row = cursor.fetchone()
            filename = table_id.identifier(grammar) + ".tsv"
            z.writestr(filename, tsv.encode("utf-8"))
        z.close()
        return memfile.getvalue()

    def get_xlsx_binary(self) -> bytes:
        """
        Other notes:
        - cell size
          http://stackoverflow.com/questions/13197574/python-openpyxl-column-width-size-adjust
          ... and the "auto_size" / "bestFit" options don't really do the job,
              according to the interweb
        """  # noqa
        wb = Workbook()
        wb.remove_sheet(wb.active)  # remove the autocreated blank sheet
        sqlsheet_rows = [["Table", "SQL", "Args", "Executed_at"]]
        for table_id, sql, args in self.patient_multiquery.all_queries():
            sqlsheet_rows.append([str(table_id), sql, repr(args),
                                  datetime.datetime.now()])
            ws = wb.create_sheet(title=str(table_id))
            with self.get_executed_cursor(sql, args) as cursor:
                fieldnames = get_fieldnames_from_cursor(cursor)
                ws.append(fieldnames)
                row = cursor.fetchone()
                while row is not None:
                    ws.append(tuple(row))
                    row = cursor.fetchone()
        sql_ws = wb.create_sheet(title="SQL")
        for r in sqlsheet_rows:
            sql_ws.append(r)
        return excel_to_bytes(wb)

    # -------------------------------------------------------------------------
    # Using the internal PatientMultiQuery
    # -------------------------------------------------------------------------

    def get_patient_id_query(self, with_order_by: bool = True) -> str:
        return self.patient_multiquery.patient_id_query(
            with_order_by=with_order_by)

    # -------------------------------------------------------------------------
    # Display
    # -------------------------------------------------------------------------

    def summary_html(self) -> str:
        # Nasty hack. We want collapsing things, so we want HTML element IDs.
        # We could build the HTML table in code for the Patient Explorer
        # chooser, but I was trying to do it in Django templates.
        # However, it's not easy to pass parameters (such as an
        # HtmlElementCounter) back to Python from Django templates.
        # So we can hack it a bit:
        element_counter = HtmlElementCounter(prefix="pe_{}_".format(self.id))
        return self.patient_multiquery.summary_html(
            element_counter=element_counter)

    def has_patient_id_query(self) -> bool:
        return self.patient_multiquery.has_patient_id_query()

    def has_output_columns(self) -> bool:
        return self.patient_multiquery.has_output_columns()

    # -------------------------------------------------------------------------
    # Data finder
    # -------------------------------------------------------------------------

    def data_finder_excel(self) -> bytes:
        """
        Performs a SELECT COUNT(*)
        Returns (fieldnames, rows).
        """
        fieldnames = []
        wb = Workbook()
        wb.remove_sheet(wb.active)  # remove the autocreated blank sheet
        all_ws = wb.create_sheet("All_patients")
        sql_ws = wb.create_sheet("SQL")
        sql_ws.append(["Table", "SQL", "Args", "Executed_at"])

        for table_identifier, sql, args in \
                self.patient_multiquery.gen_data_finder_queries():
            sql_ws.append([table_identifier,
                           format_sql(sql),
                           repr(args),
                           datetime.datetime.now()])
            with self.get_executed_cursor(sql, args) as cursor:
                if not fieldnames:
                    fieldnames = get_fieldnames_from_cursor(cursor)
                    all_ws.append(fieldnames)
                row = cursor.fetchone()
                while row is not None:
                    mrid = str(row[0])
                    if mrid in wb:
                        ws = wb[mrid]
                    else:
                        ws = wb.create_sheet(mrid)
                        ws.append(fieldnames)
                    rowtuple = tuple(row)
                    ws.append(rowtuple)
                    all_ws.append(rowtuple)
                    row = cursor.fetchone()
        return excel_to_bytes(wb)


# =============================================================================
# PatientExplorer auditing class
# =============================================================================

class PatientExplorerAudit(models.Model):
    """
    Audit log for a PatientExplorer.
    """
    id = models.AutoField(primary_key=True)  # automatic
    patient_explorer = models.ForeignKey('PatientExplorer')
    when = models.DateTimeField(auto_now_add=True)
    count_only = models.BooleanField(default=False)
    n_records = models.IntegerField(default=0)
    # ... not PositiveIntegerField; SQL Server gives -1, for example
    failed = models.BooleanField(default=False)
    fail_msg = models.TextField()

    def __str__(self):
        return "<PatientExplorerAudit id={}>".format(self.id)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-17 00:43
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('research', '0005_query_sql_hash'),
    ]

    operations = [
        migrations.AddField(
            model_name='patientexplorer',
            name='pmq_hash',
            field=models.BigIntegerField(default=0, verbose_name='64-bit non-cryptographic hash of JSON of patient_multiquery'),  # noqa
            preserve_default=False,
        ),
    ]
